"""Types for hinting.

This module and its contents should usually be imported within an :py:`if TYPE_CHECKING`
block.
"""
# pragma: exclude file

from collections.abc import Hashable, Sequence
from typing import TypeAlias, TypeVar

from pint import Quantity, Unit
from xarray.core.types import Dims, InterpOptions, ScalarOrArray

from .core.attrseries import AttrSeries
from .core.key import Key, KeyLike
from .core.quantity import AnyQuantity
from .core.sparsedataarray import SparseDataArray

__all__ = [
    "AnyQuantity",
    "Dims",
    "IndexLabel",
    "InterpOptions",
    "KeyLike",
    "ScalarOrArray",
    "TKeyLike",
    "TQuantity",
    "Unit",
]

# Mirror the definition from pandas-stubs
IndexLabel: TypeAlias = Hashable | Sequence[Hashable]

#: Similar to :any:`KeyLike`, but as a variable that can be use to match function/method
#: outputs to inputs.
TKeyLike = TypeVar("TKeyLike", Key, str)

#: Similar to :any:`.AnyQuantity`, but as a variable that can be used to match function
#: /method outputs to inputs.
TQuantity = TypeVar("TQuantity", AttrSeries, SparseDataArray)

#: A string possibly containing a unit expression, or a :mod:`pint` objects with units.
UnitLike = str | Unit | Quantity



import logging
from collections.abc import Callable, Iterable, Mapping, MutableMapping, Sequence
from copy import copy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any
from warnings import warn

from genno import operator
from genno.core.computer import Computer
from genno.core.exceptions import KeyExistsError, MissingKeyError
from genno.core.key import Key, iter_keys
from genno.util import REPLACE_UNITS

log = logging.getLogger(__name__)

#: Registry of configuration section handlers.
HANDLERS: dict[str, "ConfigHandler"] = {}

#: .. deprecated:: 1.25.0
#:    Instead, use:
#:
#:    .. code-block:: python
#:
#:       from genno.config import handles, store
#:
#:       handles("section_name", False, False)(store)
#:
#: Configuration sections/keys to be stored with no action.
STORE: set[str] = set()


def configure(path: Path | str | None = None, **config):
    """Configure :mod:`.genno` globally.

    Modifies global variables that affect the behaviour of *all* Computers and
    operators. Configuration keys loaded from file are superseded by keyword arguments.
    Messages are logged at level :obj:`logging.INFO` if `config` contains unhandled
    sections.

    Parameters
    ----------
    path : pathlib.Path, optional
        Path to a configuration file in JSON or YAML format.
    **config :
        Configuration keys/sections and values.
    """
    if path:
        config["path"] = Path(path)
    # The value of fail= doesn't matter, since no Computer is given
    parse_config(None, data=config, fail="raise")


def _convert_deprecated_store_global():
    if len(STORE):
        warn(
            'genno.config.STORE; use @handles("section_name", False, False)(store)',
            DeprecationWarning,
            stacklevel=2,
        )

        for name in STORE:
            handles(name, iterate=False, discard=False)(store)

        STORE.clear()


@dataclass
class ConfigHandler:
    """Class for a configuration key/section handler."""

    #: Configuration key or section handled.
    key: str

    #: Callable
    callback: Callable[[Computer | None, Mapping | Sequence], Any]

    #: If :any:`True`, apply :attr:`callback` iteratively to each member of the
    #: value/section.
    iterate: bool

    #: If :any:`True`, discard the configuration contents after handling.
    discard: bool

    def handle(self, data: Mapping | Sequence, c: Computer | None):
        if self.iterate:
            if isinstance(data, Mapping):
                iterator: Iterable = data.items()
            elif isinstance(data, Sequence):
                iterator = data
            else:  # pragma: no cover
                raise NotImplementedError(type(data))
            yield from [
                (("apply", self.callback), dict(info=item)) for item in iterator
            ]
        else:
            self.callback(c, data)


class PathHandler(ConfigHandler):
    """Special :class:`ConfigHandler` that reads from a YAML file."""

    def __init__(self):
        pass

    def handle(self, data, c):
        path = data.pop("path", None)
        if path is None:
            return data

        import yaml

        # Load configuration from file
        path = Path(path)
        with open(path, "r") as f:
            new_data = yaml.safe_load(f)

        # Overwrite the file content with direct configuration values
        new_data.update(data)
        data = new_data

        # Also store the directory where the configuration file was located
        data.update(config_dir=path.parent)

        return data


def handles(section_name: str, iterate: bool = True, discard: bool = True):
    """Decorator to register a configuration section handler in :data:`HANDLERS`.

    Parameters
    ----------
    section_name: str
        The name of the configuration section to handle. Using a name already present
        in :data:`HANDLERS` overrides that handler.
    iterate : bool, optional
        If :obj:`True`, the handler is called once for each item (either list item, or
        (key, value) tuple) in the section. If :obj:`False`, the entire section
        contents, of whatever type, are passed to tha handler.
    discard : bool, optional
        If :obj:`True`, configuration section data is discarded after the handler is
        called. If :obj:`False`, the data is retained and stored on the Computer.
    """

    def wrapper(f: Callable):
        ch = ConfigHandler(section_name, f, iterate, discard)

        try:
            log.debug(f"Override {HANDLERS[ch.key]!r}")
        except KeyError:
            pass
        finally:
            HANDLERS[section_name] = ch

        return f

    return wrapper


def parse_config(
    c: Computer | None,
    data: MutableMapping[str, Any],
    fail: str | int | None = None,
):
    _convert_deprecated_store_global()

    # Handle configuration from a file
    data = PathHandler().handle(data, c)

    # Assemble a queue of (args, kwargs) for Computer.add_queue()
    queue: list[tuple[tuple, dict]] = []

    # Sections to discard, e.g. with handler._store = False
    discard = set()

    for section_name, section_data in data.items():
        handler = HANDLERS.get(section_name)
        if not handler:
            log.info(f"No handler for configuration section '{section_name}:'; ignored")
            continue

        if handler.discard:
            discard.add(section_name)

        # Allow the handler to extend the queue of computations to be added
        queue.extend(handler.handle(section_data, c))

    # Discard sections so marked
    [data.pop(section_name) for section_name in discard]

    if c:
        # Store certain keys in the "config" dictionary of the graph itself
        c.graph["config"].update(data)

        # Process the entries
        c.add_queue(queue, max_tries=2, fail=fail)
    elif len(queue):
        raise RuntimeError("Cannot apply non-global configuration without a Computer")


@handles("aggregate")
def aggregate(c: Computer, info):
    """Handle one entry from the ``aggregate:`` config section."""
    # Copy for destructive .pop()
    info = copy(info)

    # Unpack `info`
    quantities = c.infer_keys(info.pop("_quantities"))
    tag = info.pop("_tag")
    # Keyword arguments for add()
    kw = dict(
        fail=info.pop("_fail", None),
        groups={info.pop("_dim"): info},
        keep=True,
        strict=True,
        sums=True,
    )

    def _log_or_raise(exc: Exception, default_level: str, message: str):
        """Either raise `exc` if ``kw["fail"]`` > `default_level`, or log `message`."""
        fail_level = getattr(logging, (kw["fail"] or default_level).upper())
        if fail_level >= logging.ERROR:
            raise exc
        else:
            log.log(fail_level, message)

    try:
        quantities = c.check_keys(*quantities)
    except MissingKeyError as e:
        # Default to fail="error" here: stricter
        _log_or_raise(e, "error", f"No key(s) {e.args!r} to aggregate")

    # Iterate over quantities to be aggregated
    for qty in map(Key, quantities):
        try:
            result = c.add(qty.add_tag(tag), "aggregate", qty, **kw)
        except KeyExistsError:
            pass
        except MissingKeyError as e:
            # Default to fail="warning": more permissive
            _log_or_raise(e, "warning", repr(e))
        else:
            if keys := list(iter_keys(result)):
                log.info(f"Add {repr(keys[0])} + {len(keys) - 1} partial sums")


@handles("alias")
def alias(c: Computer, info):
    """Handle one entry from the ``alias:`` config section."""
    c.add(info[0], info[1])


@handles("combine")
def combine(c: Computer, info):
    """Handle one entry from the ``combine:`` config section."""
    # Split inputs into three lists
    quantities, select, weights = [], [], []

    # Key for the new quantity
    key = Key(info["key"])

    # Loop over inputs to the combination
    for i in info["inputs"]:
        # Required dimensions for this input: output key's dims, plus any
        # dims that must be selected on
        selector = i.get("select", {})
        dims = set(key.dims) | set(selector.keys())
        quantities.append(c.infer_keys(i["quantity"], dims))

        select.append(selector)
        weights.append(i.get("weight", 1))

    # Check for malformed input
    assert len(quantities) == len(select) == len(weights)

    # Computation
    task = tuple(
        [partial(operator.combine, select=select, weights=weights)] + quantities
    )

    added = iter_keys(c.add(key, task, strict=True, sums=True))

    log.info(f"Add {repr(key)} + {len(list(added)) - 1} partial sums")
    log.debug("    as combination of")
    log.debug(f"    {repr(quantities)}")


@handles("default", iterate=False)
def default(c: Computer, info):
    """Handle the ``default:`` config section."""
    c.default_key = info


@handles("files")
def files(c: Computer, info):
    """Handle one entry from the ``files:`` config section."""
    # Files with exogenous data
    path = Path(info["path"])
    if not path.is_absolute():
        # Resolve relative paths relative to the directory containing the configuration
        # file
        path = c.graph["config"].get("config_dir", Path.cwd()) / path

    info["path"] = path

    c.add("load_file", **info)


@handles("general")
def general(c: Computer, info):
    """Handle one entry from the ``general:`` config section."""
    # Inputs
    # TODO allow to specify a more narrow key and *not* have infer_keys applied; perhaps
    #      using "*"
    inputs = c.infer_keys(info.get("inputs", []))

    if info["comp"] in ("mul", "product"):
        result = c.add(info["key"], "mul", *inputs)
        log.info(f"Add {repr(result)} using .add_product()")
    else:
        # The resulting key
        key = info["key"]
        key = key if Key.bare_name(key) else Key(key)

        # Infer the dimensions of the resulting key if ":*:" is given for the dims
        if isinstance(key, Key) and key.dims == ("*",):
            key = Key.product(key.name, *inputs, tag=key.tag)
            # log.debug(f"Inferred dimensions ({', '.join(key.dims)}) for '*'")

        # If info["comp"] is None, the task is a list that collects other keys
        _seq: type = list
        task = []

        if info["comp"] is not None:
            _seq = tuple  # Task is a computation
            # Retrieve the function for the computation
            f = c.get_operator(info["comp"])
            if f is None:
                raise ValueError(info["comp"])
            task = [partial(f, **info.get("args", {}))]

            log.info(f"Add {repr(key)} using {f.__module__}.{f.__name__}(…)")

        task.extend(inputs)

        added = c.add(key, _seq(task), strict=True, sums=info.get("sums", False))

        if isinstance(added, tuple):
            log.info(f"    + {len(added) - 1} partial sums")


@handles("report")
def report(c: Computer, info):
    """Handle one entry from the ``report:`` config section."""
    log.info(f"Add report {info['key']} with {len(info['members'])} table(s)")

    # Concatenate pyam data structures
    c.add(info["key"], tuple([c.get_operator("concat")] + info["members"]), strict=True)


@handles("cache_path", iterate=False, discard=False)
@handles("cache_skip", iterate=False, discard=False)
@handles("config_dir", iterate=False, discard=False)
def store(c: Computer, info):
    """Config sections/keys to be stored with no modification."""
    pass


@handles("units", iterate=False)
def units(c: Computer, info):
    """Handle the ``units:`` config section."""
    import pint

    # Define units
    registry = pint.get_application_registry()
    try:
        defs = info["define"].strip()
        registry.define(defs)
    except KeyError:
        pass
    except (TypeError, pint.DefinitionSyntaxError, pint.RedefinitionError) as e:
        log.warning(e)
    else:
        log.info(f"Apply global unit definitions {defs}")

    # Add replacements
    for old, new in info.get("replace", {}).items():
        log.info(f"Replace unit {repr(old)} with {repr(new)}")
        REPLACE_UNITS[old] = new



"""Resolve missing references using aliased target names, domains, and/or types.

Expanded from and with thanks for https://stackoverflow.com/a/62301461.
"""

import re
from collections.abc import Mapping
from typing import TYPE_CHECKING

from docutils.nodes import Text
from sphinx.addnodes import pending_xref
from sphinx.ext.intersphinx import missing_reference

if TYPE_CHECKING:
    import sphinx.application


class Replacement:
    refdomain: str | None
    reftype: str | None
    reftarget: str
    text: str | None

    # Identifier characters
    c = "[^`<>]+"

    # Match any of:
    # - ":domain:type:`text <target>`"
    # - ":domain:type:`target`"
    # - ":type:`text <target>`"
    # - ":type:`target`"
    # - "text <target>"
    # - "target"
    _target_expr = re.compile(
        rf"(:((?P<rd>{c}):)?(?P<rt>{c}):)?(`?)(?P<t_or_t>{c})(<(?P<target>{c})>)?\5"
    )

    def __init__(self, value: str) -> None:
        match = self._target_expr.fullmatch(value)
        assert match is not None
        self.refdomain, self.reftype, target_or_text, target = [
            match.group(k) for k in ("rd", "rt", "t_or_t", "target")
        ]
        if target is None:  # Target only, no replacement text
            self.reftarget, self.text = target_or_text, None
        else:  # Both target and text replacement
            self.reftarget, self.text = target, target_or_text.rstrip()


def apply_alias(config: Mapping[str, str], node) -> bool:
    """Apply `config` to `node`."""
    try:
        # Identify an alias expression matching the "reftarget" attribute of `node`
        expr = next(filter(lambda e: re.match(e, node["reftarget"]), config))
    except (KeyError, StopIteration):
        # No such attribute, or no matching expression → nothing to do
        return False

    # Unpack information about the replacement
    replace = Replacement(config[expr])

    # Resolve the ref by substituting the reftarget for the matching part
    node["reftarget"] = re.sub(f"^{expr}", replace.reftarget, node["reftarget"])

    # Rewrite the rendered text, reftype, and refdomain
    if replace.text:
        # Find the text node child
        text_node = next(iter(node.traverse(lambda n: n.tagname == "#text")))
        # Remove the old text node, add new text node with custom text
        text_node.parent.replace(text_node, Text(replace.text))
        # Force further processing to preserve this text node
        node["refexplicit"] = True
    if replace.reftype:
        node["reftype"] = replace.reftype
    if replace.refdomain:
        node["refdomain"] = replace.refdomain

    return True


def resolve_internal_aliases(app: "sphinx.application.Sphinx", doctree):
    """Handler for 'doctree-read' events."""
    config = app.config["reference_aliases"]
    for node in doctree.traverse(condition=pending_xref):
        apply_alias(config, node)


def resolve_intersphinx_aliases(app, env, node, contnode):
    """Handler for 'missing-reference' (intersphinx) events."""
    if apply_alias(app.config["reference_aliases"], node):
        # Delegate the rest of the work to intersphinx
        return missing_reference(app, env, node, contnode)


def setup(app: "sphinx.application.Sphinx") -> dict:
    """Connect :mod:`.rewrite_refs` event handlers."""

    app.add_config_value("reference_aliases", dict(), "")

    app.connect("doctree-read", resolve_internal_aliases)
    app.connect("missing-reference", resolve_intersphinx_aliases)

    return dict(parallel_read_safe=True, parallel_write_safe=True)



__all__ = [
    "codelist_to_groups",
]


def __getattr__(name: str):
    if name == "codelist_to_groups":
        from warnings import warn

        warn(
            f"Import {name} from genno.compat.sdmx; use genno.compat.sdmx.operator or "
            'Computer.require_compat("sdmx") instead',
            FutureWarning,
        )

        from . import operator

        return operator.codelist_to_groups
    else:
        raise AttributeError



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sdmx.format import Version
    from sdmx.model.common import (
        BaseDataSet,
        BaseDataStructureDefinition,
        BaseObservation,
        DimensionComponent,
        MaintainableArtefact,
    )


def handle_od(
    value: "str | DimensionComponent | None", structure: "BaseDataStructureDefinition"
) -> "DimensionComponent | None":
    """Handle `observation_dimension` arguments for :mod:`.sdmx.operator`.

    Ensure either None or a DimensionComponent.
    """
    import sdmx

    if isinstance(value, sdmx.model.common.DimensionComponent) or value is None:
        return value
    elif value is not None:
        return structure.dimensions.get(value)


def urn(obj: "MaintainableArtefact") -> str:
    """Return the URN of `obj`, or construct it."""
    import sdmx.urn

    if result := obj.urn:  # pragma: no cover
        return result
    else:
        return sdmx.urn.make(obj)


def handle_version(
    version: "str |Version | None",
) -> tuple["Version", type["BaseDataSet"], type["BaseObservation"]]:
    """Handle `version` arguments for :mod:`.sdmx.operator`.

    Also return either :mod:`sdmx.model.v21` or :mod:`sdmx.model.v30`, as appropriate.
    """
    import sdmx.model
    from sdmx.format import Version

    # Ensure a Version enum member
    if not isinstance(version, Version):
        version = Version[version or "2.1"]

    # Retrieve information model module
    im = {Version["2.1"]: sdmx.model.v21, Version["3.0.0"]: sdmx.model.v30}[version]

    return (
        version,
        im.get_class("StructureSpecificDataSet"),
        im.get_class("Observation"),
    )



from typing import TYPE_CHECKING

from sphinx.ext.autodoc import FunctionDocumenter

from genno.core.operator import Operator

if TYPE_CHECKING:
    import sphinx.application


class OperatorDocumenter(FunctionDocumenter):
    @classmethod
    def can_document_member(cls, member, membername, isattr, parent) -> bool:
        return isinstance(member, Operator) or super().can_document_member(
            member, membername, isattr, parent
        )


def setup(app: "sphinx.application.Sphinx") -> dict:
    """Configure :mod:`sphinx.ext.autodoc` to handle :class:`Operator` as functions."""
    app.add_autodocumenter(OperatorDocumenter, override=True)

    return dict(parallel_read_safe=True, parallel_write_safe=True)



from collections.abc import Callable
from functools import update_wrapper
from inspect import signature
from typing import Any, ClassVar
from warnings import warn

from .computer import Computer
from .key import KeyLike


class Operator:
    """Base class for a callable with convenience methods.

    Example
    -------
    >>> from genno import Operator
    >>>
    >>> @Operator.define()
    ... def myfunc(q1: Quantity, q2: Quantity) -> Quantity:
    ...     # Operator code
    >>>
    >>> @myfunc.helper
    ... def add_myfunc(f, computer, *args, **kwargs):
    ...     # Custom code to add tasks to `computer`
    ...     # Perform checks or handle `args` and `kwargs`.

    Or:

    >>> from genno import Operator
    >>>
    >>> def add_myfunc(f, computer, *args, **kwargs):
    ...     # ... as above
    >>>
    >>> @Operator.define(helper=add_myfunc)
    ... def myfunc(q1: Quantity, q2: Quantity) -> Quantity:
    ...     # ... as above
    """

    # Use these specific attribute names to be intelligible to functools.partial()
    #: Function or callable for the Operator.
    func: ClassVar[Callable]
    args = ()
    keywords: dict[str, Any] = dict()

    #: Helper method to add tasks to a :class:`.Computer`. Register with :meth:`helper`,
    #: invoke with :meth:`add_tasks`.
    _add_tasks: ClassVar[Callable | None] = None

    def __call__(self, *args, **kwargs):
        # The callable is stored as a static method; `self` is not passed
        return self.func(*args, **kwargs)

    def __hash__(self):
        return hash(self.func)

    def __eq__(self, other):
        """Compares equal to the wrapped `func`."""
        return other == self.func

    def __repr__(self):
        return f"<operator {self.__class__.__name__}>"

    @staticmethod
    def define(
        deprecated_func_arg: Callable | None = None,
        *,
        helper: Callable | None = None,
    ) -> Callable[[Callable], "Operator"]:
        """Return a decorator that wraps `func` in a :class:`.Operator` instance.

        Parameters
        ----------
        helper : Callable, optional
            Equivalent to calling :meth:`helper` on the Operator instance.
        """

        def decorator(func: Callable) -> "Operator":
            # This follows the pattern of using a metaclass, except compressed

            # - Create the class
            #   - Same name as func.
            #   - Subclass of Operator.
            #   - func is a static method.
            #   - Signature of klass.__call__ is the signature of func.
            klass = type(
                func.__name__,
                (Operator,),
                {"func": staticmethod(func), "__signature__": signature(func)},
            )

            # Create an instance of the class, update __doc__ and other attributes,
            # return
            # NB these are updated on the instance, not on `klass`, to satisfy Sphinx,
            #    which will skip documenting items that have the same __doc__ as their
            #    class
            result = update_wrapper(klass(), func, updated=())
            assert isinstance(result, Operator)  # For mypy

            if helper:
                # Register the provided `helper` method for the class
                result.helper(helper)

            return result

        if deprecated_func_arg is not None:
            warn(
                "@Operator.define must be called: @Operator.define()",
                DeprecationWarning,
                2,
            )
            return decorator(deprecated_func_arg)

        return decorator

    def helper(self, func: Callable[..., KeyLike | tuple[KeyLike, ...]]) -> Callable:
        """Register `func` as the convenience method for adding task(s)."""
        self.__class__._add_tasks = staticmethod(func)
        return func

    def add_tasks(self, c: "Computer", *args, **kwargs) -> tuple[KeyLike, ...]:
        """Invoke :attr:`_add_task` to add tasks to `c`."""
        if self._add_tasks is None:
            raise NotImplementedError

        return self._add_tasks(self.func, c, *args, **kwargs)



"""Tests of AttrSeries in particular."""

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from genno import Computer
from genno.core.attrseries import AttrSeries
from genno.testing import add_large_data


class TestAttrSeries:
    @pytest.fixture
    def foo(self):
        idx = pd.MultiIndex.from_product([["a1", "a2"], ["b1", "b2"]], names=["a", "b"])
        yield AttrSeries([0, 1, 2, 3], index=idx, name="Foo", units="kg")

    @pytest.fixture
    def bar(self):
        """A 1-dimensional quantity."""
        yield AttrSeries([0, 1], index=pd.Index(["a1", "a2"], name="a"))

    def test_align_levels(self, foo, bar):
        # Scalar vs scalar
        q = AttrSeries(0.1)
        other = AttrSeries(2.2)
        _, result = q.align_levels(other)
        assert tuple() == result.dims

        # Scalar to 1D: broadcasts to 1D
        _, result = q.align_levels(bar)
        assert ("a",) == result.dims
        assert 1 == len(result.unique())

        # Scalar to 2D:
        _, result = q.align_levels(foo)
        assert ("b",) == result.dims
        assert 1 == len(result.unique())

        # 1D to scalar
        _, result = bar.align_levels(q)
        assert ("a",) == result.dims

        # 2D vs. 2D; same dimensions, different order
        idx = pd.MultiIndex.from_product([["b1", "b2"], ["a1", "a2"]], names=["b", "a"])
        q = AttrSeries([3, 2, 1, 0], index=idx)
        assert ("b", "a") == q.dims
        _, result = q.align_levels(foo)
        assert ("a", "b") == result.dims

    def test_clip(self, foo: AttrSeries) -> None:
        with pytest.raises(NotImplementedError, match="keep_attrs=False"):
            foo.clip(1.0, keep_attrs=False)

    def test_cumprod(self, foo, bar):
        """AttrSeries.cumprod works with 1-dimensional quantities."""
        result0 = (1.1 + bar).cumprod("a")
        assert ("a",) == result0.dims

        # Same result with dim=None
        result1 = (1.1 + bar).cumprod()
        pdt.assert_series_equal(result0, result1)

        # But not with ≥1D
        with pytest.raises(NotImplementedError):
            foo.cumprod()

    def test_data(self, foo):
        assert isinstance(foo.data, np.ndarray)

    def test_expand_dims(self, ureg, foo):
        # Name and units pass through expand_dims
        result = foo.expand_dims(c=["c1", "c2"])
        assert foo.name == result.name and ureg.Unit("kg") == foo.units == result.units

    def test_interp(self, foo):
        with pytest.raises(NotImplementedError):
            foo.interp(coords=dict(a=["a1", "a1.5", "a2"], b=["b1", "b1.5", "b2"]))

    def test_rename(self, foo):
        assert foo.rename({"a": "c", "b": "d"}).dims == ("c", "d")

    def test_repr(self, foo):
        assert (
            "a   b \n"
            """a1  b1    0
    b2    1
a2  b1    2
    b2    3
Name: Foo, dtype: int64, units: kilogram"""
        ) == repr(foo)

    @pytest.mark.parametrize(
        "indexers_kwargs, dims", ((dict(a="a2"), ()), (dict(a=["a2"]), ("a",)))
    )
    def test_sel(self, bar, indexers_kwargs, dims) -> None:
        # Selecting 1 element from 1-D parameter still returns AttrSeries
        result = bar.sel(**indexers_kwargs)
        assert isinstance(result, AttrSeries)
        assert result.size == 1
        assert dims == result.dims
        assert result.iloc[0] == 1

    def test_sel_not_implemented(self, bar):
        with pytest.raises(NotImplementedError):
            bar.sel(a="a2", method="bfill")

        with pytest.raises(NotImplementedError):
            bar.sel(a="a2", tolerance=0.01)

    def test_shift(self, foo):
        foo.shift(a=1)
        foo.shift(b=1)

    def test_squeeze(self, foo) -> None:
        """:meth:`squeeze` results in MultiIndex.

        https://github.com/khaeru/genno/issues/120
        https://github.com/iiasa/message_ix/issues/788
        """
        # Squeeze the length-1 dimension "a"
        result0 = foo.sel(a=["a1"]).squeeze()

        # Result is 1-D but multi-indexed
        assert 1 == len(result0.dims)
        assert isinstance(result0.index, pd.MultiIndex)

        # Squeeze only 1 of >1 dimensions with length
        result1 = foo.sel(a=["a1"], b=["b2"]).squeeze("a")
        assert 1 == len(result1.dims)

        # Squeeze both
        result2 = foo.sel(a=["a1"], b=["b2"]).squeeze()
        assert 0 == len(result2.dims)
        result3 = foo.sel(a=["a1"], b=["b2"]).squeeze(dim=["a", "b"])
        assert 0 == len(result3.dims)

    def test_sum(self, foo, bar):
        # AttrSeries can be summed across all dimensions
        result = foo.sum(dim=["a", "b"])
        assert isinstance(result, AttrSeries)  # returns an AttrSeries
        assert result.size == 1  # with one element
        assert result.item() == 6  # that has the correct value

        # Sum with wrong dim raises ValueError
        with pytest.raises(ValueError):
            bar.sum("b")

        # Index with duplicate entries
        _baz = pd.DataFrame(
            [
                ["a1", "b1", "c1", 1.0],
                ["a1", "b1", "c1", 2.0],
                ["a2", "b2", "c3", 3.0],
                ["a2", "b2", "c4", 4.0],
                ["a3", "b3", "c5", 5.0],
            ],
            columns=["a", "b", "c", "value"],
        )
        # Fails with v1.13.0 AttrSeries.sum() using unstack()
        AttrSeries(_baz.set_index(["a", "b", "c"])["value"]).sum(dim="c")

        with pytest.raises(NotImplementedError):
            bar.sum("a", skipna=False)

    def test_others(self, foo, bar):
        # Exercise other compatibility functions
        assert type(foo.to_frame()) is pd.DataFrame
        assert foo.drop("a").dims == ("b",)
        assert bar.dims == ("a",)

        with pytest.raises(NotImplementedError):
            bar.item("a2")
        with pytest.raises(ValueError):
            bar.item()

    def test_where(self, foo: AttrSeries) -> None:
        with pytest.raises(NotImplementedError, match="drop=True"):
            foo.where(lambda x: x, drop=True)
        with pytest.raises(NotImplementedError, match="axis=….* inplace=…"):
            foo.where(lambda x: x, axis=1)
        with pytest.raises(NotImplementedError, match="axis=….* inplace=…"):
            foo.where(lambda x: x, inplace=True)


@pytest.mark.skip(reason="Slow, for benchmarking only")
def test_sum_large(N_data=1e7):  # pragma: no cover
    """Test :meth:`.AttrSeries.sum` for large, sparse data."""
    # Create a single large AttrSeries
    c = Computer()
    add_large_data(c, 1, N_dims=11, N_data=N_data)
    qty = c.get("q_00")

    # Compute a sum()
    result = qty.sum(dim=["j", "k"])
    # print(result)  # DEBUG
    del result



import logging
import re
from importlib import import_module

import pytest

from genno import Computer, Key, MissingKeyError, config, configure
from genno.compat.pyam import HAS_PYAM
from genno.config import HANDLERS, ConfigHandler, handles


def test_deprecated_store_global(recwarn):
    config.STORE.add("FOO")

    # Warning is emitted here since Computer.__init__() calls parse_config()
    with pytest.warns(DeprecationWarning, match="genno.config.STORE"):
        c = Computer()

    # STORE was emptied, converted to handler in HANDLERS
    assert 0 == len(config.STORE)

    # Value is handled correctly
    c.configure(FOO="BAR")
    assert "BAR" == c.graph["config"]["FOO"]

    # Restore state for other tests
    HANDLERS.pop("FOO")


def test_handlers():
    # Account for handlers in packages modules that may be installed in the user's
    # testing environment
    third_party_handlers = 0
    try:
        import_module("ixmp")
    except ImportError:
        pass
    else:  # pragma: no cover
        third_party_handlers += 2

    # Expected config handlers are available
    assert 11 + (1 * HAS_PYAM) + third_party_handlers == len(HANDLERS)

    # Handlers are all callable
    for key, ch in HANDLERS.items():
        assert isinstance(key, str) and isinstance(ch, ConfigHandler)


@pytest.mark.parametrize(
    "name",
    [
        "config-aggregate0.yaml",
        "config-aggregate1.yaml",
        pytest.param(
            "config-aggregate2.yaml", marks=pytest.mark.xfail(raises=MissingKeyError)
        ),
        "config-combine.yaml",
        "config-general0.yaml",
        pytest.param(
            "config-general1.yaml", marks=pytest.mark.xfail(raises=ValueError)
        ),
        "config-report.yaml",
        "config-units.yaml",
    ],
)
def test_file(test_data_path, name):
    """Test handling configuration file syntax using test data files."""
    c = Computer()

    # Set up test contents
    c.add(Key("X", list("abc")), None, sums=True)
    c.add(Key("Y", list("bcd")), None, sums=True)

    c.configure(path=test_data_path / name)


def test_general_infer_dims():
    """Test dimension inference in handling "general:" config items."""
    c = Computer()

    # Set up test contents
    c.add(Key("X", list("abcd")), None, sums=True)
    c.add(Key("Y", list("cefg")), None, sums=True)

    c.configure(general=[dict(comp="concat", key="Z:*:foo", inputs=["X", "Y"])])

    # Dimensions were inferred
    assert "Z:a-b-c-d-e-f-g:foo" in c


def test_global(test_data_path):
    configure(path=test_data_path / "config-units.yaml")

    with pytest.raises(
        RuntimeError, match="Cannot apply non-global configuration without a Computer"
    ):
        configure(path=test_data_path / "config-global.yaml")


def test_handles(caplog, monkeypatch):
    """:func:`handles` raises a warning when used twice."""
    monkeypatch.delitem(HANDLERS, "foo", raising=False)
    caplog.set_level(logging.DEBUG)

    @handles("foo")
    def foo1(c: Computer, info):
        """Test function; never executed."""

    assert len(caplog.messages) == 0

    @handles("foo")
    def foo2(c: Computer, info):
        """Test function; never executed."""

    assert 1 == len(caplog.messages)
    assert re.match(
        r"Override ConfigHandler\(key='foo', "
        "callback=<function test_handles.<locals>.foo1 [^>]*>, iterate=True, "
        r"discard=True\)",
        caplog.messages[0],
    )



import re
from typing import cast

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from xarray.testing import assert_equal as assert_xr_equal

import genno
from genno import Computer
from genno.core.sparsedataarray import SparseDataArray
from genno.operator import random_qty
from genno.testing import add_test_data

sparse = pytest.importorskip(
    "sparse",
    reason="`sparse` not available → can't test SparseDataArray",
)


class TestSparseAccessor:
    @pytest.fixture()
    def x(self):
        x_series = pd.Series(
            data=[1.0, 2, 3, 4],
            index=pd.MultiIndex.from_product(
                [["a", "b"], ["c", "d"]], names=["foo", "bar"]
            ),
        )
        yield SparseDataArray.from_series(x_series)

    @pytest.fixture()
    def y(self):
        y_series = pd.Series(data=[5.0, 6], index=pd.Index(["e", "f"], name="baz"))
        yield SparseDataArray.from_series(y_series)

    def test_all(self, x: SparseDataArray, y: SparseDataArray):
        """Test conversion to sparse.COO-backed xr.DataArray."""
        x_dense = x._sda.dense_super
        y_dense = y._sda.dense_super
        assert not x_dense._sda.COO_data or x_dense._sda.nan_fill
        assert not y_dense._sda.COO_data or y_dense._sda.nan_fill

        # As of sparse 0.10, sparse `y` is automatically broadcast to `x_dense`
        # Previously, this raised ValueError.
        x_dense * y

        z1 = x_dense._sda.convert() * y

        z2 = x * y_dense._sda.convert()
        assert z1.dims == ("foo", "bar", "baz") == z2.dims
        assert_xr_equal(z1, z2)

        z3 = x._sda.convert() * y._sda.convert()
        assert_xr_equal(z1, z3)

        z4 = x._sda.convert() * y
        assert_xr_equal(z1, z4)

        z5 = x * y
        assert_xr_equal(z1, z5)

    def test_dense(self):
        # Trying to densify data that is already-dense raises a TypeError handled by
        # SparseAccessor
        z = SparseDataArray(1.0)
        assert 1.0 == z._sda.dense._sda.dense.item()

        # Same with empty/null data
        z = SparseDataArray()
        assert np.isnan(z._sda.dense._sda.dense.item())


@pytest.mark.usefixtures("quantity_is_sparsedataarray")
class TestSparseDataArray:
    def test_init(self, caplog) -> None:
        # SDA can be initialized with integer data; a warning is logged
        SparseDataArray([[0, 1], [2, 3]])
        assert any(re.match(r"Force dtype int\w+ → float", m) for m in caplog.messages)

        # SDA can be initialized with sparse.COO with fill_value other than np.nan; this
        # is replaced with nan
        data = sparse.COO([], fill_value=0.0)
        assert not np.isnan(data.fill_value)
        q = SparseDataArray(data)
        assert np.isnan(q.data.fill_value)

    def test_item(self) -> None:
        # Works on a multi-dimensional quantity
        q = random_qty(dict(x=9, y=9, z=9))
        assert 0 <= q.sel(x="x8", y="y8", z="z8").item() <= 1

        with pytest.raises(ValueError, match="can only convert an array of size 1"):
            q.item()

        with pytest.raises(NotImplementedError):
            q.item(1, 2, 3)

    def test_loc(self) -> None:
        """SparseDataArray.loc[] accessor works.

        For some version prior to sparse 0.11.2, a specific workaround was required, but
        no longer. Retain the test to catch any regression.
        """
        *_, x = add_test_data(Computer())

        # .loc accessor works, returns same class as object but 1 element
        assert isinstance(x.loc["foo1", 2040], SparseDataArray)
        assert isinstance(x.loc["foo1", 2040].item(), float)

    def test_scalar(self) -> None:
        """Scalar Quantities can be created."""
        A = genno.Quantity(1.0, units="kg")
        B = genno.Quantity(2.0, units="kg")

        # Fragment occurring in .operator.add()
        list(map(genno.Quantity, xr.broadcast(*cast(xr.DataArray, (A, B)))))



_warned = False


def __getattr__(name):
    global _warned
    if not _warned:
        from warnings import warn

        warn(
            f"Importing from {__name__} will be deprecated in a future version; "
            "use genno.operator instead.",
            FutureWarning,
            2,
        )
        _warned = True

    from . import operator

    return getattr(operator, name)



import sys
from typing import TYPE_CHECKING, Literal, Type

from .attrseries import AttrSeries
from .base import BaseQuantity
from .sparsedataarray import SparseDataArray

if TYPE_CHECKING:
    # TODO Remove this block once Python 3.10 is the lowest supported version
    from typing import TypeAlias


def assert_quantity(*args):
    """Assert that each of `args` is a Quantity object.

    Raises
    ------
    TypeError
        with a indicative message.
    """
    for i, arg in enumerate(args):
        if not isinstance(arg, BaseQuantity):
            raise TypeError(
                f"arg #{i + 1} ({repr(arg)[:20]}) is not Quantity; likely an incorrect key"
            )


def get_class() -> Type[AttrSeries | SparseDataArray]:
    """Get the current :class:`.Quantity` implementation in use.

    Returns one of the classes :class:`.AttrSeries` or :class:`.SparseDataArray`.
    """
    global Quantity
    return Quantity


def set_class(
    name: Literal["AttrSeries", "SparseDataArray"] = "AttrSeries",
) -> Type[AttrSeries | SparseDataArray]:
    """Set the :class:`.Quantity` implementation to be used.

    This also updates :py:`genno.Quantity` and :py:`genno.quantity.Quantity` to refer to
    the selected class. It does **not** update previously-imported references to one
    class or the other; code that uses :func:`.set_class` should refer to one of those
    two locations:

    .. code-block::

       import genno
       from genno import Quantity  # AttrSeries, by default

       Quantity()        # AttrSeries
       genno.Quantity()  # AttrSeries

       genno.set_class("SparseDataArray")

       Quantity()        # AttrSeries
       genno.Quantity()  # SparseDataArray

    Another approach is to update the local reference with the return value of the
    function:

    .. code-block:: python

       from genno import Quantity, set_class

       Quantity()  # AttrSeries

       Quantity = set_class("SparseDataArray")

       Quantity()  # SparseDataArray

    In code that does not use :func:`.set_class`, :py:`from genno import Quantity` is
    safe.

    See also
    --------
    .AnyQuantity
    """
    global CLASS

    try:
        cls = {"AttrSeries": AttrSeries, "SparseDataArray": SparseDataArray}[name]
    except KeyError:
        raise ValueError(f"no Quantity implementation {name}")

    # Update globals in the current module and at the top level
    for module in "genno.core.quantity", "genno":
        setattr(sys.modules[module], "Quantity", cls)

    # Update global
    CLASS = cls.__name__

    return cls


#: Class used to implement :class:`.Quantity`.
Quantity: "TypeAlias" = AttrSeries

#: Name of :class:`.Quantity`.
CLASS = "AttrSeries"

#: Either :class:`.AttrSeries` or :class:`.SparseDataArray`. Code in :mod:`genno` or
#: user code that receives or returns any Quantity implementation should be typed with
#: this type.
AnyQuantity = AttrSeries | SparseDataArray



"""Tests for genno.quantity."""

import logging
import operator
import re
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pint
import pytest
import xarray as xr
from numpy import nan
from pytest import param

import genno
import genno.operator
from genno import Computer
from genno.core.attrseries import AttrSeries
from genno.core.quantity import assert_quantity, get_class, set_class
from genno.core.sparsedataarray import SparseDataArray
from genno.testing import (
    MARK,
    add_large_data,
    assert_qty_allclose,
    assert_qty_equal,
    assert_units,
)

if TYPE_CHECKING:
    import genno.core.base
    from genno.core.quantity import AnyQuantity

pytestmark = pytest.mark.usefixtures("parametrize_quantity_class")

SUPPORTED_BINOPS = [
    operator.add,
    operator.mul,
    operator.pow,
    operator.sub,
    operator.truediv,
]


class TestQuantity:
    """Tests of Quantity."""

    @pytest.fixture
    def a(self):
        yield genno.Quantity([0.8, 0.2], coords={"p": ["oil", "water"]})

    @pytest.fixture
    def foo(self):
        # NB 0.0 because sparse cannot handle data that is all int
        yield genno.Quantity(
            [[0.0, 1], [2, 3]],
            coords={"a": ["a1", "a2"], "b": ["b1", "b2"]},
            name="Foo",
            units="kg",
        )

    @pytest.fixture()
    def tri(self):
        """Fixture returning triangular data to test fill, shift, etc."""
        return genno.Quantity(
            xr.DataArray(
                [
                    [nan, nan, 1.0, nan, nan],
                    [nan, 2, 3, 4, nan],
                    [5, 6, 7, 8, 9],
                ],
                coords=[
                    ("x", ["x0", "x1", "x2"]),
                    ("y", ["y0", "y1", "y2", "y3", "y4"]),
                ],
            ),
            units="kg",
        )

    @pytest.mark.parametrize(
        "args, kwargs",
        (
            # Integer, converted to float() for sparse
            ((3,), dict(units="kg")),
            # Scalar object
            ((object(),), dict(units="kg")),
            # pd.Series
            ((pd.Series([0, 1], index=["a", "b"], name="foo"),), dict(units="kg")),
            # pd.DataFrame
            (
                (pd.DataFrame([[0], [1]], index=["a", "b"], columns=["foo"]),),
                dict(units="kg"),
            ),
            # xr.DataArray, 0-D
            (xr.DataArray([3.0]), dict()),
            # xr.DataArray, 2-D
            (
                xr.DataArray(
                    [[0.0, 1], [2, 3]], coords={"a": ["a1", "a2"], "b": ["b1", "b2"]}
                ),
                dict(),
            ),
            # xarray-style arguments
            (
                ([[0.0, 1], [2, 3]],),  # float
                dict(coords={"a": ["a1", "a2"], "b": ["b1", "b2"]}),
            ),
            (
                ([[0, 1], [2, 3]],),  # int
                dict(coords={"a": ["a1", "a2"], "b": ["b1", "b2"]}),
            ),
            pytest.param(
                (
                    pd.DataFrame(
                        [[0, 1], [2, 3]], index=["a", "b"], columns=["foo", "bar"]
                    ),
                ),
                dict(units="kg"),
                marks=pytest.mark.xfail(raises=TypeError),
            ),
        ),
    )
    def test_init(self, args, kwargs) -> None:
        """Instantiated from a scalar object."""
        genno.Quantity(*args, **kwargs)

    def test_assert(self, a) -> None:
        """Test assertions about Quantity.

        These are tests without `attr` property, in which case direct pd.Series
        and xr.DataArray comparisons are possible.
        """
        with pytest.raises(
            TypeError,
            match=re.escape("arg #2 ('foo') is not Quantity; likely an incorrect key"),
        ):
            assert_quantity(a, "foo")

        # Convert to pd.Series
        b = a.to_series()

        assert_qty_equal(a, b, check_type=False)
        assert_qty_equal(b, a, check_type=False)
        assert_qty_allclose(a, b, check_type=False)
        assert_qty_allclose(b, a, check_type=False)

        c = genno.Quantity(a)

        assert_qty_equal(a, c, check_type=True)
        assert_qty_equal(c, a, check_type=True)
        assert_qty_allclose(a, c, check_type=True)
        assert_qty_allclose(c, a, check_type=True)

    def test_assert_with_attrs(self, a) -> None:
        """Test assertions about Quantity with attrs.

        Here direct pd.Series and xr.DataArray comparisons are *not* possible.
        """
        attrs = {"foo": "bar"}
        a.attrs = attrs

        b = genno.Quantity(a)

        # make sure it has the correct property
        assert a.attrs == attrs
        assert b.attrs == attrs

        assert_qty_equal(a, b)
        assert_qty_equal(b, a)
        assert_qty_allclose(a, b)
        assert_qty_allclose(b, a)

        # check_attrs=False allows a successful equals assertion even when the
        # attrs are different
        a.attrs = {"bar": "foo"}
        assert_qty_equal(a, b, check_attrs=False)

    def test_assign_coords(self, a) -> None:
        # Relabel an existing dimension
        q1 = a.assign_coords({"p": ["apple", "orange"]})
        assert ("p",) == q1.dims
        assert all(["apple", "orange"] == q1.coords["p"])

        # Exception raised when the values are of the wrong length
        with pytest.raises(
            ValueError,
            # NB "and" with xarray <= 2025.4; "but" with xarray >= 2025.6
            match="conflicting sizes for dimension 'p': length 2 .* (but|and) length 3",
        ):
            a.assign_coords({"p": ["apple", "orange", "banana"]})
        with pytest.raises(
            ValueError,
            match="conflicting sizes for dimension 'p': length 2 .* (but|and) length 1",
        ):
            a.assign_coords({"p": ["apple"]})

    def test_astype(self, tri) -> None:
        result = tri.astype(float)
        assert float == result.dtype

    def test_bfill(self, tri) -> None:
        """Test Quantity.bfill()."""
        if genno.Quantity is SparseDataArray:
            pytest.xfail(reason="sparse.COO.flip() not implemented")

        r1 = tri.bfill("x")
        assert r1.loc["x0", "y0"] == tri.loc["x2", "y0"]

        r2 = tri.bfill("y")
        assert r2.loc["x0", "y0"] == tri.loc["x0", "y2"]

    def test_clip(self, tri) -> None:
        # Only min=
        assert 1.0 == tri.loc["x0", "y2"].item()
        r1 = tri.clip(2.0)
        assert 2.0 == r1.loc["x0", "y2"].item()

        # Only max=
        assert 9.0 == tri.loc["x2", "y4"].item()
        r2 = tri.clip(max=8.0)
        assert 8.0 == r2.loc["x2", "y4"].item()

        # Both min= and max=
        r3 = tri.clip(2.0, 8.0)
        assert 2.0 == r3.loc["x0", "y2"].item()
        assert 8.0 == r3.loc["x2", "y4"].item()

    def test_coords(self, tri) -> None:
        coords = tri.coords
        assert isinstance(coords, xr.core.coordinates.Coordinates)
        assert ["x", "y"] == list(coords)
        assert "x" in coords  # __contains__

        assert isinstance(coords["x"], xr.DataArray)

        coords = genno.Quantity(3, units="kg").coords
        assert [] == list(coords)

    def test_copy_modify(self, a) -> None:
        """Making a Quantity another produces a distinct attrs dictionary."""
        assert 0 == len(a.attrs)

        a.units = pint.Unit("km")

        b = genno.Quantity(a, units="kg")
        assert pint.Unit("kg") == b.units

        assert pint.Unit("km") == a.units

    def test_cumprod(self, caplog, tri) -> None:
        """Test Quantity.cumprod()."""
        if genno.Quantity is SparseDataArray:
            pytest.xfail(reason="sparse.COO.nancumprod() not implemented")

        caplog.set_level(logging.INFO)

        args = dict(axis=123) if genno.Quantity is AttrSeries else dict()
        r1 = tri.cumprod("x", **args)
        assert 1 * 3 * 7 == r1.loc["x2", "y2"]
        if genno.Quantity is AttrSeries:
            assert ["AttrSeries.cumprod(…, axis=…) is ignored"] == caplog.messages

        r2 = tri.cumprod("y")
        assert 2 * 3 == r2.loc["x1", "y2"]
        assert 5 * 6 * 7 * 8 * 9 == r2.loc["x2", "y4"]

    def test_drop_vars(self, a) -> None:
        a.expand_dims({"phase": ["liquid"]}).drop_vars("phase")

    def test_expand_dims(self, a) -> None:
        # Single label on a new dimension
        q0 = a.expand_dims({"phase": ["liquid"]})
        assert ("phase", "p") == q0.dims

        # New dimension(s) without labels
        q1 = a.expand_dims(["phase"])
        assert ("phase", "p") == q1.dims
        assert 2 == q1.size
        assert (1, 2) == q1.shape

        # New dimension(s) without labels
        q2 = a.expand_dims({"phase": []})
        assert ("phase", "p") == q2.dims
        if genno.Quantity is AttrSeries:
            # NB this behaviour differs slightly from xr.DataArray.expand_dims()
            assert (1, 2) == q2.shape
            assert 2 == q2.size
        else:
            # da = xr.DataArray([0.8, 0.2], coords=[["oil", "water"]], dims=["p"])
            # assert (0, 2) == da.expand_dims({"phase": []}).shape  # Different result
            # assert (1, 2) == da.expand_dims(["phase"]).shape  # Same result

            assert (0, 2) == q2.shape
            assert 0 == q2.size

        # Multiple labels
        q3 = a.expand_dims({"phase": ["liquid", "solid"]})
        assert ("phase", "p") == q3.dims
        assert all(["liquid", "solid"] == q3.coords["phase"])

        # Multiple dimensions and labels
        q4 = a.expand_dims({"colour": ["red", "blue"], "phase": ["liquid", "solid"]})
        assert ("colour", "phase", "p") == q4.dims

    def test_ffill(self, tri) -> None:
        """Test Quantity.ffill()."""

        # Forward fill along "x" dimension results in no change
        r1 = tri.ffill("x")
        assert_qty_equal(tri, r1)

        # Forward fill along y dimension works
        r2 = tri.ffill("y")

        # Check some filled values
        assert (
            r2.loc["x0", "y4"].item()
            == r2.loc["x0", "y3"].item()
            == tri.loc["x0", "y2"].item()
        )

    def test_keep(self, foo, tri) -> None:
        assert {"_unit"} == set(foo.attrs) == set(tri.attrs)

        # Assign attributes to foo
        foo.attrs["bar"] = "baz"

        # Use foo._keep to preserve attributes on `tri`
        result = foo._keep(tri, attrs=True)
        assert result is tri

        # Attributes from `foo` pass through
        assert_units(result, "kg")
        assert "baz" == result.attrs["bar"]

        # Now assign new attrs via _keep()
        result = foo._keep(tri, attrs={"bar": "qux"}, units=True)
        assert_units(result, "kg")
        assert "qux" == result.attrs["bar"]

    @pytest.mark.parametrize(
        "left, right", (["float", "qty"], ["qty", "float"], ["qty", "qty"])
    )
    @pytest.mark.parametrize("op", SUPPORTED_BINOPS)
    def test_operation(self, left, op, right, tri: "AnyQuantity") -> None:
        """Test the standard binary operations +, -, *, /."""
        values = {
            "float": 1.0,
            "qty": genno.operator.assign_units(tri, "dimensionless"),
        }
        left = values[left]
        right = values[right]

        # Binary operation succeeds
        result = op(left, right)

        # Result is of the expected type
        assert isinstance(result, tri.__class__), type(result)

    @pytest.mark.parametrize("op", SUPPORTED_BINOPS)
    @pytest.mark.parametrize("type_", [int, float, param(str, marks=pytest.mark.xfail)])
    def test_operation_scalar(self, op, type_, a) -> None:
        """Quantity can be added to int or float."""
        result = op(type_(4.2), a)

        # Result has the expected shape
        assert (2,) == result.shape
        assert a.dtype == result.dtype

    @pytest.mark.parametrize(
        "op, left_units, right_units, exp_units",
        (
            (operator.add, "", "", ""),  # Both dimensionless
            (operator.add, "kg", "kg", "kg"),  # Same units
            pytest.param(
                operator.add,
                "",
                "kg",
                "",
                marks=pytest.mark.xfail(raises=pint.DimensionalityError),
            ),
            (operator.sub, "", "", ""),  # Both dimensionless
            (operator.sub, "kg", "kg", "kg"),  # Same units
            pytest.param(
                operator.sub,
                "",
                "kg",
                "",
                marks=pytest.mark.xfail(raises=pint.DimensionalityError),
            ),
            (operator.mul, "", "", ""),  # Both dimensionless
            (operator.mul, "", "kg", "kg"),  # One dimensionless
            (operator.mul, "kg", "kg", "kg **2"),  # Same units
            (operator.mul, "kg", "litre", "kg * litre"),  # Different units
            (operator.truediv, "", "", ""),  # Both dimensionless
            (operator.truediv, "kg", "", "kg"),  # Denominator dimensionless
            (operator.truediv, "", "kg", "1 / kg"),  # Numerator dimensionless
            (operator.truediv, "kg", "kg", ""),  # Same units
            (operator.truediv, "kg", "litre", "kg / litre"),  # Different units
        ),
    )
    def test_operation_units0(
        self, a: "AnyQuantity", op, left_units, right_units, exp_units
    ) -> None:
        """Test units pass through the binary operations between Quantities."""
        left = genno.Quantity(a, units=left_units)
        right = genno.Quantity(a, units=right_units)

        # Binary operation succeeds
        result = op(left, right)

        # Result is of the expected type
        assert isinstance(result, a.__class__), type(result)

        # Result has the expected units
        assert_units(result, exp_units)

    @pytest.mark.parametrize(
        "op, side, units_in, exp_units",
        (
            (operator.add, "L", "", ""),
            (operator.add, "R", "", ""),
            (operator.add, "L", "kg", False),
            (operator.add, "R", "kg", False),
            (operator.mul, "L", "", ""),
            (operator.mul, "R", "", ""),
            (operator.mul, "L", "kg", "kg"),
            (operator.mul, "R", "kg", "kg"),
            (operator.pow, "L", "", ""),
            (operator.pow, "R", "", ""),
            (operator.pow, "L", "kg", "kg ** 2"),
            (operator.pow, "R", "kg", False),
            (operator.sub, "L", "", ""),
            (operator.sub, "R", "", ""),
            (operator.sub, "L", "kg", False),
            (operator.sub, "R", "kg", False),
            (operator.truediv, "L", "", ""),
            (operator.truediv, "R", "", ""),
            (operator.truediv, "L", "kg", "kg"),
            (operator.truediv, "R", "kg", "1 / kg"),
        ),
    )
    def test_operation_units1(
        self, a: "AnyQuantity", op, side, units_in, exp_units
    ) -> None:
        """Test units pass through the binary operations between Quantity and scalar."""
        from contextlib import nullcontext

        q = genno.Quantity(a, units=units_in)
        other = 2.0

        left, right = (q, other) if side == "L" else (other, q)

        # Binary operation succeeds
        with pytest.raises(Exception) if exp_units is False else nullcontext():
            result = op(left, right)

        if exp_units is not False:
            # Result is of the expected type
            assert isinstance(result, a.__class__), type(result)

            # Result has the expected units
            assert_units(result, exp_units)

    def test_pipe(self, ureg, tri) -> None:
        result = tri.pipe(genno.operator.assign_units, "km")
        assert ureg.Unit("km") == result.units

    @pytest.mark.parametrize(
        "args, dropped",
        (
            (dict(x="x1"), True),  # default drop=False
            (dict(x=["x1"]), False),  # default drop=False
            (dict(x="x1", drop=False), True),
            (dict(x=["x1"], drop=False), False),
            (dict(x="x1", drop=True), True),
            (dict(x=["x1"], drop=True), False),
        ),
    )
    def test_sel(self, tri, args, dropped) -> None:
        result = tri.sel(**args)

        assert ({"y"} if dropped else {"x", "y"}) == set(result.dims)

    @MARK["issue/145"]
    def test_sel_xarray(self, tri) -> None:
        """xarray-style indexing works."""
        # Create indexers
        newdim = {"newdim": ["nd0", "nd1", "nd2"]}
        x_idx = xr.DataArray(["x2", "x1", "x2"], coords=newdim)
        y_idx = xr.DataArray(["y4", "y2", "y0"], coords=newdim)

        # Select using the indexers
        # NB with pandas 2.1, this triggers the RecursionError fixed in khaeru/genno#99
        assert_qty_equal(
            genno.Quantity([9.0, 3.0, 5.0], coords=newdim, units="kg"),
            tri.sel(x=x_idx, y=y_idx),
            ignore_extra_coords=True,
        )

        # Exception raised for mismatched lengths
        with pytest.raises(IndexError, match="Dimensions of indexers mismatch"):
            tri.sel(x=x_idx[:-1], y=y_idx)

    def test_shift(self, tri) -> None:
        """Test Quantity.shift()."""
        if genno.Quantity is SparseDataArray:
            pytest.xfail(reason="sparse.COO.pad() not implemented")

        r1 = tri.shift(x=1)
        assert r1.loc["x2", "y1"] == tri.loc["x1", "y1"]

        r2 = tri.shift(y=2)
        assert r2.loc["x2", "y4"] == tri.loc["x2", "y2"]

        r3 = tri.shift(x=1, y=2)
        assert r3.loc["x2", "y4"] == tri.loc["x1", "y2"]

    def test_size(self) -> None:
        """Stress-test reporting of large, sparse quantities."""
        # Create the Reporter
        c = Computer()

        # Prepare large data, store the keys of the quantities
        keys = add_large_data(c, num_params=10)

        # Add a task to compute the product, i.e. requires all the q_*
        c.add("bigmem", tuple([genno.operator.mul] + keys))

        # One quantity fits in memory
        c.get(keys[0])

        if genno.Quantity is SparseDataArray:
            pytest.xfail(
                reason='"IndexError: Only one-dimensional iterable indices supported." '
                "in sparse._coo.indexing"
            )

        # All quantities can be multiplied without raising MemoryError
        result = c.get("bigmem")

        # Result can be converted to pd.Series
        result.to_series()

    @pytest.mark.parametrize(
        "sel_kw, dims, values",
        (
            (dict(a=["a1"]), ("b",), [0, 1]),
            (dict(a="a1"), ("b",), [0, 1]),
            (dict(a="a2", b="b1"), (), [2]),  ####
            (dict(a=["a2"], b="b1"), (), [2]),
            (dict(a=["a2"], b=["b1"]), (), [2]),
        ),
    )
    def test_squeeze0(self, foo: SparseDataArray, sel_kw, dims, values) -> None:
        # Method succeeds
        result = foo.sel(**sel_kw).squeeze()

        # Dimensions as expected
        assert dims == result.dims

        # Values as expected
        pdt.assert_series_equal(
            pd.Series(values, name="Foo"),
            result.to_series(),
            check_series_type=False,
            check_index=False,
            check_dtype=False,
        )

    @pytest.mark.parametrize(
        "dim, exc_type, match",
        (
            (
                "b",
                ValueError,
                "dimension to squeeze out which has length greater than one",
            ),
            ("c", KeyError, "c"),
        ),
    )
    def test_squeeze1(self, foo, dim, exc_type, match) -> None:
        with pytest.raises(exc_type, match=match):
            print(foo.squeeze(dim=dim))

    def test_sum(self, foo) -> None:
        """:meth:`.sum` handles :any:`Ellipsis`."""
        assert_qty_equal(
            foo.sum(["a", "b"]),
            foo.sum(...),
        )

    def test_to_dataframe(self, a) -> None:
        """Test Quantity.to_dataframe()."""
        # Returns pd.DataFrame
        result = a.to_dataframe()
        assert isinstance(result, pd.DataFrame)

        # "value" is used as a column name
        assert ["value"] == result.columns

        # Explicitly passed name produces a named column
        assert ["foo"] == a.to_dataframe("foo").columns

        with pytest.raises(NotImplementedError):
            a.to_dataframe(dim_order=["foo", "bar"])

    def test_to_series(self, a) -> None:
        """Test .to_series() on child classes, and Quantity.from_series."""
        s = a.to_series()
        assert isinstance(s, pd.Series)

        genno.Quantity.from_series(s)

    def test_units(self, a: "AnyQuantity") -> None:
        # Units can be retrieved; dimensionless by default
        assert a.units.dimensionless

        # Set with a string results in a pint.Unit instance
        a.units = "kg"
        assert pint.Unit("kg") == a.units

        # Can be set to dimensionless
        a.units = ""
        assert a.units.dimensionless  # type: ignore [attr-defined]

    def test_where(self, tri: "AnyQuantity") -> None:
        assert np.isnan(tri.sel(x="x0", y="y0").item())

        # cond=Callable, other=scalar
        value = 8.8
        q0 = tri.where(np.isfinite, value)
        assert value == q0.sel(x="x0", y="y0").item()

        # cond=lambda function, other=scalar
        value = 0.1
        q1 = tri.where(lambda v: v % 2 != 0, 0.1)
        assert value == q1.sel(x="x2", y="y1").item()


@pytest.mark.parametrize(
    "value, cls",
    (
        ("AttrSeries", AttrSeries),
        ("SparseDataArray", SparseDataArray),
        pytest.param("FOO", None, marks=pytest.mark.xfail(raises=ValueError)),
    ),
)
def test_set_class(value, cls):
    import genno as g1
    import genno.core.quantity as gcq1

    set_class(value)

    import genno as g2
    import genno.core.quantity as gcq2

    assert (
        g1.Quantity
        is gcq1.Quantity
        is g2.Quantity
        is gcq2.Quantity
        is get_class()
        is cls
    )



"""Compatibility with :mod:`dask`."""


def cull(dsk, keys):
    """Like :func:`dask.optimization.cull`.

    This version calls :func:`.to_keylike` on the culled graph and dependencies to
    ensure the resulting graph does not contain :class:`.Key`.
    """
    import dask.optimization

    out0, dependencies0 = dask.optimization.cull(dsk, keys)

    # Rewrite Key to str in out0
    out1 = {to_keylike(k): to_keylike(task) for k, task in out0.items()}

    # Rewrite Key to str in dependencies0
    dependencies1 = {to_keylike(k): to_keylike(deps) for k, deps in out0.items()}

    return out1, dependencies1


def to_keylike(value):
    """Rewrite :class:`.Key` `value` (or in `value`) to :class:`.str`.

    Collections such as :class:`tuple` and :class:`list` are rewritten recursively.
    """
    from genno.core.key import Key

    if isinstance(value, (str, bytes, int, float)):
        # These are the strict types of dask.typing.Key (also tuple of same).
        # Return as-is without further checks
        return value
    elif isinstance(value, Key):
        return str(value)
    elif type(value) in (tuple, list):  # NB Only exactly these types; not subclasses
        return type(value)(map(to_keylike, value))  # Recurse; return same type
    else:
        return value



import logging
import types
from collections import deque
from collections.abc import (
    Callable,
    Hashable,
    Iterable,
    Mapping,
    MutableSequence,
    Sequence,
)
from copy import copy
from functools import lru_cache, partial
from importlib import import_module
from inspect import signature
from itertools import compress
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast
from warnings import catch_warnings, warn

import dask
import pint
import xarray as xr
from dask.core import quote

from genno import caching, operator
from genno.compat.dask import cull
from genno.compat.xarray import either_dict_or_kwargs
from genno.util import partial_split, update_recursive

from .describe import describe_recursive
from .exceptions import ComputationError, KeyExistsError, MissingKeyError
from .graph import Graph
from .key import Key

if TYPE_CHECKING:
    import genno.core.graph
    import genno.core.key
    from genno.core.key import KeyLike
    from genno.types import TKeyLike


log = logging.getLogger(__name__)

#: Emit :class:`.FutureWarning` from :meth:`.Computer.add` when :class:`.tuple` is
#: returned. This default value can be overridden with
#: :py:`c.configure(config={"warn on result tuple": False})`.
DEFAULT_WARN_ON_RESULT_TUPLE = False


class Computer:
    """Class for describing and executing computations.

    Parameters
    ----------
    kwargs :
        Passed to :meth:`configure`.
    """

    #: A dask-format graph (see :doc:`1 <dask:graphs>`, :doc:`2 <dask:spec>`).
    graph: "genno.core.graph.Graph" = Graph(config=dict())

    #: The default key to :meth:`.get` with no argument.
    default_key: "KeyLike | None" = None

    #: List of modules containing operators.
    #:
    #: By default, this includes the :mod:`genno` built-in operators in
    #: :mod:`genno.operator`. :meth:`require_compat` appends additional modules,
    #: for instance :mod:`genno.compat.plotnine`, to this list. User code may also add
    #: modules to this list directly.
    modules: MutableSequence[types.ModuleType] = []

    # Action to take on failed items on add_queue(). This is a stack; the rightmost
    # element is current; the leftmost is the default.
    _queue_fail: MutableSequence[int]

    def __init__(self, **kwargs):
        self.graph = Graph(config=dict())
        self.modules = [operator]
        self._queue_fail = deque([logging.ERROR])
        self.configure(**kwargs)

    # Python data model

    def __contains__(self, item) -> bool:
        return self.graph.__contains__(item)

    def __ior__(self, other: "Computer") -> "Computer":
        """Same as :meth:`.update`."""
        self.update(other)
        return self

    def __or__(self, other: "Computer") -> "Computer":
        """Return a new Computer with the union of the contents of two Computers."""
        result = Computer()
        result |= self
        result |= other
        return result

    def __setitem__(self, data: "KeyLike", *args) -> None:
        _args, kwargs = args[0], {}

        if isinstance(_args[-1], dict):
            *_args, kwargs = _args

        self.add(data, *_args, **kwargs)

    # Dask data model

    def __dask_keys__(self):
        return self.graph.keys()

    def __dask_graph__(self):
        return self.graph

    # Configuration

    def configure(
        self,
        path: Path | str | None = None,
        fail: str | int = "raise",
        config: Mapping[str, Any] | None = None,
        **config_kw,
    ):
        """Configure the Computer.

        Accepts a `path` to a configuration file and/or keyword arguments.
        Configuration keys loaded from file are superseded by keyword arguments.
        Messages are logged at level :data:`logging.INFO` if `config` contains
        unhandled sections.

        See :doc:`config` for a list of all configuration sections and keys, and details
        of the configuration file format.

        Parameters
        ----------
        path : pathlib.Path, optional
            Path to a configuration file in JSON or YAML format.
        fail : "raise" or str or :mod:`logging` level, optional
            Passed to :meth:`.add_queue`. If not "raise", then log messages are
            generated for config handlers that fail. The Computer may be only partially
            configured.
        config :
            Configuration keys/sections and values, as a mapping. Use this if any of
            the keys/sections are not valid Python names, for instance if they contain
            "-" or " ".

        Other parameters
        ----------------
        **config_kw :
            Configuration keys/sections and values, as keyword arguments.
        """
        from genno.config import parse_config

        config = {
            str(k): v
            for k, v in either_dict_or_kwargs(config, config_kw, "configure").items()
        }
        if path:
            if "path" in config:
                raise ValueError('cannot give both path= and a "path" key in config=…')
            config.setdefault("path", Path(path))

        parse_config(self, data=config, fail=fail)

    # Manipulating callables

    def get_operator(self, name) -> Callable | None:
        """Return a function, :class:`.Operator`, or callable for use in a task.

        :meth:`get_operator` checks each of the :attr:`modules` for a callable with the
        given `name`. Modules at the end of the list take precedence over those earlier
        in the list.

        Returns
        -------
        callable
        None
            If there is no callable with the given `name` in any of :attr:`modules`.
        """
        if not isinstance(name, str):
            # `name` is not a string; can't be the name of a function/class/object
            return None

        # Cached call with `name` guaranteed to be hashable
        return self._get_operator(name)

    @lru_cache()
    def _get_operator(self, name: str) -> Callable | None:
        for module in reversed(self.modules):
            try:
                # Retrieve the operator from `module`
                with catch_warnings(record=True) as cw:
                    result = getattr(module, name)
            except AttributeError:
                continue  # `name` not in this module
            else:
                if len(cw) and any(wm.category is DeprecationWarning for wm in cw):
                    # TODO Re-emit any non-DeprecationWarning
                    continue  # Some DeprecationWarning raised; don't use this import
                else:
                    return result

        # Nothing found
        return None

    #: Alias of :meth:`get_operator`.
    get_comp = get_operator

    def require_compat(self, pkg: str | types.ModuleType):
        """Register a module for :meth:`get_operator`.

        The specified module is appended to :attr:`modules`.

        Parameters
        ----------
        pkg : str or module
            One of:

            - the name of a package (for instance "plotnine"), corresponding to a
              submodule of :mod:`genno.compat` (:mod:`genno.compat.plotnine`).
              ``genno.compat.{pkg}.operator`` is added.
            - the name of any importable module, for instance "foo.bar".
            - a module object that has already been imported.

        Raises
        ------
        ModuleNotFoundError
            If the required packages are missing.

        Examples
        --------
        Operators packaged with genno for compatibility:

        >>> c = Computer()
        >>> c.require_compat("pyam")

        Operators in another module, using the module name:

        >>> c.require_compat("ixmp.reporting.computations")

        or using imported module object directly:

        >>> import ixmp.reporting.computations as mod
        >>> c.require_compat(mod)

        """
        if isinstance(pkg, types.ModuleType):
            mod = pkg
        elif "." in pkg:
            mod = import_module(pkg)
        else:
            name = f"genno.compat.{pkg}"
            # Check the upstream/third-party package is available
            if not getattr(import_module(name), f"HAS_{pkg.upper()}"):
                raise ModuleNotFoundError(
                    f"No module named '{pkg}', required by genno.compat.{pkg}"
                )
            mod = import_module(f"{name}.operator")

        # Don't duplicate
        if mod not in self.modules:
            self.modules.append(mod)

            # Clear the lookup cache
            # TODO also clear on manual changes to self.modules
            self._get_operator.cache_clear()

    # Add computations to the Computer

    def add(self, data, *args, **kwargs) -> "KeyLike | tuple[KeyLike, ...]":
        """General-purpose method to add computations.

        :meth:`add` can be called in several ways; its behaviour depends on `data`; see
        below. It chains to methods such as :meth:`add_single`, :meth:`add_queue`,
        and/or :meth:`apply`; each can also be called directly.

        Returns
        -------
        KeyLike or tuple of KeyLike
            Some or all of the keys added to the Computer.

        See also
        ---------
        add_single
        add_queue
        apply
        .iter_keys
        .single_key
        """

        # Other methods
        if isinstance(data, Sequence) and not isinstance(data, str):
            # Sequence of (args, kwargs) or args; use add_queue()
            return _warn_on_result(self, self.add_queue(data, *args, **kwargs))
        elif isinstance(data, str) and data in dir(self) and data != "add":
            # Name of another method such as "apply" or "eval"
            return _warn_on_result(self, getattr(self, data)(*args, **kwargs))

        # Possibly identify a named or direct callable in `data` or `args[0]`
        func: Callable | None = None
        if func := self.get_operator(data):
            # `data` is the name of a pre-defined computation
            # NB in the future, could raise some warning here to suggest the second form
            pass
        else:
            # Further checks
            if not isinstance(data, (Key, str)):
                raise TypeError(f"{type(data)} `data` argument")
            elif not len(args):
                raise TypeError("At least 1 argument required")

            # Check if the first element of `args` references a computation is callable
            func = self.get_operator(args[0]) or (
                args[0] if callable(args[0]) else None
            )

            # Located a callable in args[0], so `data` joins args[1:]
            if func:
                args = (data,) + args[1:]

        if func:
            try:
                # Use an implementation of Operator.add_tasks()
                return _warn_on_result(
                    self,
                    func.add_tasks(self, *args, **kwargs),  # type: ignore [attr-defined]
                )
            except (AttributeError, NotImplementedError):
                # Operator obj that doesn't implement .add_tasks(), or plain callable
                _partialed_func, kw = partial_split(func, kwargs)
                key = args[0]
                computation = (_partialed_func,) + args[1:]
        else:
            # `func` is None, for instance args[0] is a list of keys to be collected, or
            # some literal value
            key = data
            computation = args
            kw = kwargs

        # Keyword arguments not understood by .add_single() and/or .add_queue()
        sums = kw.pop("sums", False)
        fail = kw.pop("fail", "fail")

        # Add a single computation
        result = self.add_single(key, *computation, **kw)

        # Optionally add sums
        if isinstance(result, Key) and sums:
            # Add one entry for each of the partial sums of `result`
            return _warn_on_result(
                self, (result,) + self.add_queue(result.iter_sums(), fail=fail)
            )
        else:
            # NB This might be deprecated to simplify expectations of calling code
            return _warn_on_result(self, result)

    def cache(self, func):
        """Decorate `func` so that its return value is cached.

        See also
        --------
        :doc:`cache`
        """
        return caching.decorate(func, computer=self)

    def add_queue(  # noqa: C901  FIXME reduce complexity from 11 → ≤10
        self,
        queue: Iterable[tuple],
        max_tries: int = 1,
        fail: str | int | None = None,
    ) -> tuple["KeyLike", ...]:
        """Add tasks from a list or `queue`.

        Parameters
        ----------
        queue : iterable of tuple
            Each item is either a N-:class:`tuple` of positional arguments to
            :meth:`add`, or a 2-:class:`tuple` of (:class:`.tuple` of positional
            arguments, :class:`dict` of keyword arguments).
        max_tries : int, optional
            Retry adding elements up to this many times.
        fail : "raise" or str or :mod:`logging` level, optional
            Action to take when a computation from `queue` cannot be added after
            `max_tries`: "raise" an exception, or log messages on the indicated level
            and continue.
        """
        # Determine the action (log level and/or raise exception) when queue items fail
        if isinstance(fail, str):
            # Convert a string like 'debug' to logging.DEBUG
            fail = cast(int, getattr(logging, fail.upper(), logging.ERROR))
        elif fail is None:
            fail = self._queue_fail[-1]  # Use the same value as an outer call.

        # Accumulate added keys
        added: list["KeyLike"] = []

        class Item:
            """Container for queue items."""

            def __init__(self, value):
                self.count = 1
                if (
                    len(value) == 2
                    and isinstance(value[0], tuple)
                    and isinstance(value[1], Mapping)
                ):
                    self.args, self.kwargs = value  # Both args and kwargs provided
                else:
                    self.args, self.kwargs = value, {}  # `value` is positional only

        def _log(msg: str, i: Item, e: Exception | None = None, level=logging.DEBUG):
            """Log information for debugging."""
            log.log(
                level,
                f"{msg.format(i)} (max {max_tries}):\n    ({repr(i.args)}, "
                f"{repr(i.kwargs)})" + (f"\n    with {repr(e)}" if e else ""),
            )

        # Iterate over elements from queue, then any which are re-appended to be
        # retried. On the first pass, count == 1; on subsequent passes, it is
        # incremented.
        _queue = deque(map(Item, queue))
        while len(_queue):
            item = _queue.popleft()
            self._queue_fail.append(fail)

            try:
                # Recurse
                keys = self.add(*item.args, **item.kwargs)
            except KeyError as exc:
                # Adding failed
                if item.count < max_tries:
                    # This may only be due to items being out of order; retry silently
                    item.count += 1
                    _queue.append(item)

                    # verbose; uncomment for debugging only
                    # _log("Failed {0.count} times, will retry", item, exc)
                else:
                    # Failed `max_tries` times; something has gone wrong
                    _log("Failed {0.count} time(s), discarded", item, exc, fail)
                    if fail >= logging.ERROR:
                        raise  # Also raise
            else:
                # Succeeded; record the key(s)
                added.extend(keys) if isinstance(keys, tuple) else added.append(keys)

                # verbose; uncomment for debugging only
                # if count > 1:
                #     _log("Succeeded on {0.count} try", item)
            finally:
                # Restore the failure action from an outer level
                self._queue_fail.pop()

        return tuple(added)

    # Generic graph manipulations
    def add_single(
        self, key: "KeyLike", *computation, strict=False, index=False
    ) -> "KeyLike":
        """Add a single `computation` at `key`.

        Parameters
        ----------
        key : str or Key or hashable
            A string, Key, or other value identifying the output of `computation`.
        computation : object
            Any computation. See :attr:`graph`.
        strict : bool, optional
            If True, `key` must not already exist in the Computer, and any keys
            referred to by `computation` must exist.
        index : bool, optional
            If True, `key` is added to the index as a full-resolution key, so it can be
            later retrieved with :meth:`full_key`.

        Raises
        ------
        ~genno.KeyExistsError
            If `strict` is :obj:`True` and either (a) `key` already exists; or (b)
            `sums` is :obj:`True` and the key for one of the partial sums of `key`
            already exists.
        ~genno.MissingKeyError
            If `strict` is :obj:`True` and any key referred to by `computation` does
            not exist.
        """
        # Unpack a length-1 tuple, except for a tuple starting with a callable (task
        # with no arguments)
        if len(computation) == 1 and (
            isinstance(computation[0], Key) or not callable(computation[0])
        ):
            computation = computation[0]

        if index:
            warn(
                "add_single(…, index=True); full keys are automatically indexed",
                DeprecationWarning,
            )

        key = Key.bare_name(key) or Key(key)

        if strict:
            if key in self.graph:
                raise KeyExistsError(key)

            # Check valid keys in `computation` and maybe rewrite
            computation = self._rewrite_comp(computation)

        # Add to the graph
        self.graph[key] = computation

        return key

    def _rewrite_comp(self, computation):
        """Check and rewrite `computation`.

        If `computation` is :class:`tuple` or :class:`list`, it may contain other keys
        that :mod:`dask` must locate in the :attr:`graph`. Check these using
        :meth:`check_keys`, and return a modified `computation` with these in exactly
        the form they appear in the graph. This ensures dask can locate them for
        :meth:`get` and :meth:`describe`.
        """
        if not isinstance(computation, (list, tuple)):
            # Something else, such as pd.DataFrame or a literal
            return computation

        # Assemble the result using either checked keys (with properly ordered
        # dimensions) or unmodified elements from `computation`; cast to the same type
        return type(computation)(
            self.check_keys(
                *computation, predicate=lambda e: not isinstance(e, (Key, str))
            )
        )

    def apply(
        self, generator: Callable, *keys, **kwargs
    ) -> "KeyLike | tuple[KeyLike, ...]":
        """Add computations by applying `generator` to `keys`.

        Parameters
        ----------
        generator : callable
            Function to apply to `keys`. This function **may** take a first positional
            argument annotated with :class:`.Computer` or a subtype; if so, then it is
            provided with a reference to `self`.

            The function **may**:

            - :py:`yield` or return an iterable of (`key`, `computation`). These are
              used to directly update the :attr:`graph`, and then :meth:`.apply` returns
              the added keys.
            - If it is provided with a reference to the Computer, call :meth:`.add` or
              any other method to update the graph. In this case, it **should**
              :py:`return` a :class:`.Key` or sequence of keys, indicating what was
              added; these are in turn returned by :meth:`.apply`.
        keys : Hashable
            The starting key(s). These are provided as positional arguments to
            `generator`.
        kwargs
            Keyword arguments to `generator`.
        """
        args: list[Any] = self.check_keys(*keys)

        try:
            # Inspect the generator function
            par = signature(generator).parameters
            # Name of the first parameter
            par_0 = list(par.keys())[0]
        except IndexError:
            pass  # No parameters to generator
        else:
            a = par[par_0].annotation
            if isinstance(a, str) and a.endswith("Computer") or issubclass(a, Computer):
                # First parameter wants a reference to the Computer object
                args.insert(0, self)

        # Call the generator. Might return None, or yield some computations
        applied = generator(*args, **kwargs)

        if applied is None:
            return ()
        elif isinstance(applied, (Key, str)):
            return applied
        elif isinstance(applied, (list, tuple)) and isinstance(applied[0], (Key, str)):
            return tuple(applied)
        else:
            # Update the graph with the computations
            result = []
            for key, comp in applied:
                self.graph[key] = comp
                result.append(key)

            return tuple(result) if len(result) > 1 else result[0]

    def duplicate(self, key: "TKeyLike", tag: str) -> "TKeyLike":
        """Duplicate the task at `key` and all of its inputs.

        Re

        Parameters
        ----------
        key
            Starting key to duplicate.
        tag
            :attr:`~.Key.tag` to add to duplicated keys.
        """

        comp = self.graph[key]  # Retrieve the existing computation at `key`
        new_key = type(key)(Key(key) + tag)  # Identify the new key; same type as `key`

        if isinstance(comp, (list, tuple)):
            # Rewrite the computation
            new_comp = [self.duplicate(x, tag) if x in self.graph else x for x in comp]
            self.graph[new_key] = type(comp)(new_comp)
        else:
            self.graph[new_key] = comp

        return new_key

    def eval(self, expr: str) -> tuple[Key, ...]:
        r"""Evaluate `expr` to add tasks and keys.

        Parse a statement or block of statements using :mod:`.ast` from the Python
        standard library. `expr` may include:

        - Constants.
        - References to existing keys in the Computer by their name; these are expanded
          using :meth:`full_key`.
        - Multiple statements on separate lines or separated by ";".
        - Python arithmetic operators including ``+``, ``-``, ``*``, ``/``, ``**``;
          these are mapped to the corresponding :mod:`.operator`.
        - Function calls, also mapped to the corresponding :mod:`.operator` via
          :meth:`get_operator`. These may include simple positional (constants or key
          references) or keyword (constants only) arguments.

        Parameters
        ----------
        expr : str
            Expression to be evaluated.

        Returns
        -------
        tuple of Key
            One key for the left-hand side of each expression.

        Raises
        ------
        NotImplementedError
            For complex expressions not supported; if any of the statements is anything
            other than a simple assignment.
        NameError
            If a function call references a non-existent computation.
        """
        from .eval import Parser

        # Parse `expr`
        p = Parser(self)
        p.parse(expr)

        # Add tasks
        self.add_queue(p.queue)

        # Return the new keys corresponding to the LHS of each expression
        return tuple(p.new_keys.values())

    def get(self, key=None):
        """Execute and return the result of the computation `key`.

        Only `key` and its dependencies are computed.

        Parameters
        ----------
        key : str, optional
            If not provided, :attr:`default_key` is used.

        Raises
        ------
        ValueError
            If `key` and :attr:`default_key` are both :obj:`None`.
        """
        if key is None:
            if self.default_key is not None:
                key = self.default_key
            else:
                raise ValueError("no default reporting key set")
        else:
            key = self.check_keys(key)[0]

        # Protect 'config' dict, so that dask schedulers do not try to interpret its
        # contents as further tasks. Workaround for
        # https://github.com/dask/dask/issues/3523
        self.graph["config"] = quote(self.graph.get("config", dict()))

        # Cull the graph, leaving only those needed to compute *key*
        dsk, _ = cull(self.graph, key)
        log.debug(f"Cull {len(self.graph)} -> {len(dsk)} keys")

        try:
            # Dask doesn't know about genno.Key; pass a str with original dim order
            result = dask.get(dsk, str(key))
        except Exception as exc:
            raise ComputationError(exc) from None
        else:
            return result
        finally:
            # Unwrap config from protection applied above
            self.graph["config"] = self.graph["config"][0].data

    def insert(self, key: "KeyLike", *args, tag: str = "pre", **kwargs) -> None:
        """Insert a task before `key`, using `args`, `kwargs`.

        The existing task at `key` is moved to :py:`key + tag`. The `args` and `kwargs`
        are passed to :meth:`add` to insert a new task at `key`. The `args` must include
        at least 2 items:

        1. the new :class:`callable` or :class:`Operator`, and
        2. the :any:`.Ellipsis` (:py:`...`), which is replaced by the shifted
           :py:`key + tag`.

        If there are more than 2 items, each instance of the :class:`.Ellipsis` is
        replaced per (2); all other items (and `kwargs`) are passed on as-is.

        The effect is that all existing tasks to which `key` are input will receive,
        instead, the output of the added task.

        One way to use :func:`insert` is with a ‘pass-through’ `operation` that, for
        instance, performs logging, assertions, or other steps, then returns its input
        unchanged. It is also possible to insert a new task that mutates its input in
        certain ways.
        """
        # Determine a key for the task to be shifted
        k_pre = self.infer_keys(key) + tag
        if k_pre in self:
            # Cannot shift `key` because the target key already exists
            raise KeyExistsError(k_pre)

        # Construct the arguments for the add() call
        if len(args) < 2:
            raise ValueError(
                "Must supply at least 2 args (operator, ...) to Computer.insert(); "
                f"got {args}"
            )
        elif Ellipsis not in args:
            raise ValueError(f"One arg must be '...'; got {args}")

        _args = [k_pre if a is Ellipsis else a for a in args]

        try:
            # Preserve the existing task at `key`
            existing = copy(self.graph[key])
            # Add `operation` at `key`, operating on the output of the original task
            self.add(key, *_args, **kwargs)
        except Exception:
            raise
        else:
            # Move the existing task at `key` to `k_pre`
            self.graph[k_pre] = existing

    # Convenience methods for the graph and its keys

    def keys(self):
        """Return the keys of :attr:`~genno.Computer.graph`."""
        return self.graph.keys()

    def full_key(self, name_or_key: "KeyLike") -> "KeyLike":
        """Return the full-dimensionality key for `name_or_key`.

        An quantity 'foo' with dimensions (a, c, n, q, x) is available in the Computer
        as ``'foo:a-c-n-q-x'``. This :class:`.Key` can be retrieved with::

            c.full_key("foo")
            c.full_key("foo:c")
            # etc.

        Raises
        ------
        KeyError
            if `name_or_key` is not in the graph.
        """
        result = self.graph.full_key(name_or_key)
        if result is None:
            raise KeyError(name_or_key)
        return result

    def check_keys(
        self, *keys: str | Key, predicate=None, action="raise"
    ) -> list["KeyLike"]:
        """Check that `keys` are in the Computer.

        Parameters
        ----------
        keys : KeyLike
            Some :class:`Keys <Key>` or strings.
        predicate : callable, optional
            Function to run on each of `keys`; see below.
        action : "raise" or str
            Action to take on missing `keys`.

        Returns
        -------
        list of KeyLike
            One item for each item ``k`` in `keys`:

            1. ``k`` itself, unchanged, if `predicate` is given and ``predicate(k)``
               returns :obj:`True`.
            2. :meth:`.Graph.unsorted_key`, that is, ``k`` but with its dimensions in a
               specific order that already appears in :attr:`graph`.
            3. :meth:`.Graph.full_key`, that is, an existing key with the name ``k``
               with its full dimensionality.
            4. :obj:`None` otherwise.

        Raises
        ------
        ~genno.MissingKeyError
            If `action` is "raise" and 1 or more of `keys` do not appear (either in
            different dimension order, or full dimensionality) in the :attr:`graph`.
        """
        # Suppress traceback from within this function
        __tracebackhide__ = True

        if predicate:
            _p = predicate
        else:
            # Default predicate: always false
            def _p(x):
                return False

        def _check(value):
            if _p(value):
                return value
            else:
                return self.graph.unsorted_key(value) or self.graph.full_key(value)

        # Process all keys to produce more useful error messages
        result = list(map(_check, keys))

        if action == "raise":
            # Construct an exception with only (non-None) `keys` that correspond to None
            # in `result`
            exc = MissingKeyError(
                *filter(None, compress(keys, map(lambda r: r is None, result)))
            )
            if len(exc.args):  # 1 or more keys missing
                raise exc

        return result

    def infer_keys(
        self, key_or_keys: "KeyLike | Iterable[KeyLike]", dims: Iterable[str] = []
    ):
        """Infer complete `key_or_keys`.

        Each return value is one of:

        - a :class:`Key` with either

          - dimensions `dims`, if any are given, otherwise
          - its full dimensionality (cf. :meth:`full_key`)

        - :class:`str`, the same as input, if the key is not defined in the Computer.

        Parameters
        ----------
        key_or_keys : KeyLike or list of KeyLike
        dims : list of str, optional
            Drop all but these dimensions from the returned key(s).

        Returns
        -------
        KeyLike
            If `key_or_keys` is a single KeyLike.
        list of KeyLike
            If `key_or_keys` is an iterable of KeyLike.
        """
        single = isinstance(key_or_keys, (Key, Hashable))
        keys = [key_or_keys] if single else tuple(cast(Iterable, key_or_keys))

        result = list(map(partial(self.graph.infer, dims=dims), keys))

        return result[0] if single else tuple(result)

    def describe(self, key=None, quiet=True):
        """Return a string describing the computations that produce `key`.

        If `key` is not provided, all keys in the Computer are described.

        Unless `quiet`, the string is also printed to the console.

        Returns
        -------
        str
            Description of computations. If a malformed :attr:`.graph` is detected (one
            key is its own direct ancestor), the text “← CYCLE DETECTED” is shown, and
            recursion stops.
        """
        # TODO accept a list of keys, like get()
        if key is None:
            # Sort with 'all' at the end
            key = tuple(
                sorted(filter(lambda k: k != "all", self.graph.keys())) + ["all"]
            )
        else:
            key = tuple(self.check_keys(key))

        result = describe_recursive(self.graph, key)
        if not quiet:
            print(result, end="\n")
        return result

    def update(self, other: "Computer") -> None:
        """Update Computer with the contents of `other`.

        The operators :py:`|` and :py:`|=` invoke this method.

        Examples
        --------
        >>> c1 = Computer()
        >>> c2 = Computer()
        ### Create a new Computer containing all tasks from both c1 and c2
        >>> c3 = c1 | c2
        ### Add all tasks from c2 to c1
        >>> c1.update(c2)
        ### Same as above
        >>> c1 |= c2

        Raises
        ------
        RuntimeError
            if any key is present in both the Computer and `other` with a different
            task.
        """
        keys_self = set(self.graph)
        keys_other = set(other.graph) - {"config"}

        # Check matching keys for conflict before update
        for k in keys_other & keys_self:
            if self.graph[k] != other.graph[k]:
                raise RuntimeError(
                    f"Existing task {k} → {self.graph[k]} would be overwritten by "
                    + repr(other.graph[k])
                )

        # Transfer non-matching keys
        for k in keys_other - keys_self:
            self.graph[k] = other.graph[k]

        # Merge configuration
        if "config" in other.graph:
            update_recursive(self.graph.setdefault("config", {}), other.graph["config"])

    def visualize(self, filename, key=None, optimize_graph=False, **kwargs):
        """Generate an image describing the Computer structure.

        This is similar to :func:`dask.visualize`; see
        :func:`.compat.graphviz.visualize`. Requires
        `graphviz <https://pypi.org/project/graphviz/>`__.
        """
        from dask.base import collections_to_dsk, unpack_collections

        from genno.compat.graphviz import visualize

        # In dask, these calls appear in dask.base.visualize; see docstring of
        # .compat.graphviz.visualize
        args, _ = unpack_collections(self, traverse=False)
        dsk = dict(collections_to_dsk(args, optimize_graph=optimize_graph))

        if key:
            # Cull the graph, leaving only those needed to compute *key*
            N = len(dsk)
            dsk, _ = cull(dsk, key)
            log.debug(f"Cull {N} -> {len(dsk)} keys")

        return visualize(dsk, filename=filename, **kwargs)

    def write(self, key, path, **kwargs):
        """Compute `key` and write the result directly to `path`."""
        # Call the method directly without adding it to the graph
        key = self.check_keys(key)[0]
        self.get_operator("write_report")(self.get(key), path, kwargs)

    @property
    def unit_registry(self):
        """The :class:`pint.UnitRegistry` used by the Computer."""
        return pint.get_application_registry()

    # Deprecated methods

    def add_file(self, *args, **kwargs):
        """Deprecated.

        .. deprecated:: 1.18.0
           Instead use :func:`.add_load_file` via:

           .. code-block:: python

              c.add(..., "load_file", ...)
        """
        arg = (args[1:2] if len(args) else None) or None
        warn(
            f"Computer.add_file(…). Use: Computer.add({kwargs.get('key', arg)!r}, "
            '"load_file", …)',
            DeprecationWarning,
            stacklevel=2,
        )
        return operator.load_file.add_tasks(self, *args, **kwargs)

    def add_product(self, *args, **kwargs):
        """Deprecated.

        .. deprecated:: 1.18.0
           Instead use :func:`.add_binop` via:

           .. code-block:: python

              c.add(..., "mul", ...)
        """
        warn(
            f'Computer.add_product(…). Use: Computer.add({args[0]!r}, "mul", …)',
            DeprecationWarning,
            stacklevel=2,
        )
        return operator.mul.add_tasks(self, *args, **kwargs)

    def aggregate(
        self,
        qty: "KeyLike",
        tag: str,
        dims_or_groups: Mapping | str | Sequence[str],
        weights: xr.DataArray | None = None,
        keep: bool = True,
        sums: bool = False,
        fail: str | int | None = None,
    ):
        """Deprecated.

        Add a computation that aggregates `qty`.

        .. deprecated:: 1.18.0

           Instead, for a mapping/:class:`dict` `dims_or_groups`, use:

           .. code-block:: python

              c.add(qty, "aggregate", groups=dims_or_groups, keep=keep, ...)

           Or, for :class:`str` or sequence of :class:`str` `dims_or_groups`, use:

           .. code-block:: python

              c.add(None, "sum", qty, dimensions=dims_or_groups, ...)

        Parameters
        ----------
        qty: :class:`Key` or str
            Key of the quantity to be aggregated.
        tag: str
            Additional string to add to the end the key for the aggregated
            quantity.
        dims_or_groups: str or iterable of str or dict
            Name(s) of the dimension(s) to sum over, or nested dict.
        weights : :class:`xarray.DataArray`, optional
            Weights for weighted aggregation.
        keep : bool, optional
            Passed to :meth:`operator.aggregate <genno.operator.aggregate>`.
        sums : bool, optional
            Passed to :meth:`add`.
        fail : str or int, optional
            Passed to :meth:`add_queue` via :meth:`add`.

        Returns
        -------
        :class:`Key`
            The key of the newly-added node.
        """
        if isinstance(dims_or_groups, dict):
            groups = dims_or_groups
            if len(groups) > 1:
                raise NotImplementedError("aggregate() along >1 dimension")

            key = Key(qty).add_tag(tag)
            args: tuple[Any, ...] = (operator.aggregate, qty, quote(groups), keep)
            kwargs = dict()

            msg = (
                f'dims_or_groups={{...}}). Use:\nComputer.add({key!r}, "aggregate", '
                f"{qty!r}, groups=dims_or_groups, strict=True, ...)"
            )
        else:
            dims = dims_or_groups
            if isinstance(dims, str):
                dims = [dims]

            key = Key(qty).drop(*dims).add_tag(tag)
            args = ("sum", qty, weights)
            kwargs = dict(dimensions=dims)

            msg = (
                f'dims_or_groups=[...]). Use:\nComputer.add(None, "sum", {qty!r}, '
                f"dimensions=dims_or_groups, strict=True, ...)"
            )

        warn(f"Computer.aggregate(…, {msg}", DeprecationWarning, stacklevel=2)

        return self.add(key, *args, **kwargs, strict=True, sums=sums, fail=fail)

    add_aggregate = aggregate

    def convert_pyam(self, *args, **kwargs):
        """Deprecated.

        .. deprecated:: 1.18.0

           Instead use :func:`.add_as_pyam` via:

           .. code-block:: python

              c.require_compat("pyam")
              c.add(..., "as_pyam", ...)
        """
        arg0 = (repr(args[0]) + ", ") if len(args) else ""
        warn(
            f"""Computer.convert_pyam(…). Use:
    Computer.require_compat("pyam")
    Computer.add({arg0}"as_pyam", …)""",
            DeprecationWarning,
            stacklevel=2,
        )
        self.require_compat("pyam")
        return self.get_operator("as_pyam").add_tasks(self, *args, **kwargs)

    def disaggregate(self, qty, new_dim, method="shares", args=[]):
        """Deprecated.

        .. deprecated:: 1.18.0

           Instead, for `method` = "disaggregate_shares", use:

           .. code-block:: python

              c = Computer()
              c.add(qty.append(new_dim), "mul", qty, ..., strict=True)

           Or for a :func:`callable` `method`, use:

           .. code-block:: python

              c.add(qty.append(new_dim), method, qty, ..., strict=True)
        """
        # Compute the new key
        key = Key(qty).append(new_dim)

        if method == "shares":
            arg0 = (repr(args[0]) + ", ") if len(args) else ""
            msg = (
                f'method="shares"). Use: Computer.add({key!r}, "mul", {qty!r}, '
                f"{arg0}…, strict=True)"
            )
            method = "mul"
        elif callable(method):
            msg = (
                f"method={method.__name__}). Use: Computer.add({key!r}, "
                f"{method.__name__}, {qty!r}, …, strict=True)"
            )
        else:
            raise ValueError(method) if isinstance(method, str) else TypeError(method)

        warn(f"Computer.disaggregate(…, {msg}", DeprecationWarning, stacklevel=2)

        return self.add(key, method, qty, *args, sums=False, strict=True)


def _warn_on_result(computer: Computer, result):
    if isinstance(result, tuple) and computer.graph.get("config", {}).get(
        "warn on result tuple", DEFAULT_WARN_ON_RESULT_TUPLE
    ):
        warn(
            f"Return {len(result)}-tuple from Computer.add(); in a future version of "
            f"genno only the first added Key ({result[0]}) will be returned",
            FutureWarning,
            stacklevel=2,
        )
    return result



import pytest

from genno import Key, Quantity
from genno.core.graph import Graph


class TestGraph:
    @pytest.fixture
    def g(self):
        g = Graph()
        g["foo:c-b-a"] = 1
        yield g

    def test_contains0(self, g) -> None:
        """__contains__ handles incompatible types, returning False."""
        q = Quantity()
        assert (q in g) is False

    def test_contains1(self, g) -> None:
        """__contains__ handles compatible types."""
        # Compare to a key originally str and unsorted
        assert ("foo:c-b-a" in g) is True
        assert (Key("foo:c-b-a") in g) is True
        assert (Key("foo:a-b-c") in g) is True

        # Compare to a key originally Key and sorted
        g[Key("bar:x-y-z")] = None
        assert ("bar:x-y-z" in g) is True
        assert ("bar:z-x-y" in g) is True
        assert (Key("bar:x-y-z") in g) is True
        assert (Key("bar:z-x-y") in g) is True

    def test_delitem(self, g) -> None:
        assert Key("foo", "cba") == g.full_key("foo")
        del g["foo:c-b-a"]
        assert None is g.full_key("foo")

    def test_getitem(self, g: Graph) -> None:
        # Item is accessible using an equivalent KeyLike with dims in different order
        g[Key("bar:x-y-z")] = 2
        assert 2 == g[Key("bar:y-z-x")]
        assert 2 == g["bar:y-z-x"]

        # …however this only works if the item was originally added as a Key
        with pytest.raises(KeyError):
            g["foo:a-b-c"]

    def test_infer(self, g) -> None:
        g["foo:x-y-z:bar"] = 2
        g["config"] = dict(baz="qux")

        # Correct result for str or Key argument
        assert "foo:c-b-a" == g.infer("foo")
        assert "foo:c-b-a" == g.infer(Key("foo"))

        # Correct result for str or Key argument with tag
        assert "foo:x-y-z:bar" == g.infer("foo::bar")
        assert "foo:x-y-z:bar" == g.infer(Key("foo", tag="bar"))

        # String passes through
        for k in ("config", "baz"):
            result = g.infer(k)
            assert isinstance(result, str) and k == result

    def test_pop(self, g) -> None:
        assert Key("foo", "cba") == g.full_key("foo")
        assert 1 == g.pop("foo:c-b-a")
        assert None is g.full_key("foo")

    def test_setitem(self, g) -> None:
        g[Key("baz", "cba")] = 2

        assert Key("baz", "cba") == g.unsorted_key(Key("baz", "cab"))

    def test_update(self, g) -> None:
        g.update([("foo:y-x", 1), ("bar:m-n", 2)])
        assert Key("bar", "mn") == g.full_key("bar")

        g.update(baz=3)
        assert Key("baz") == g.unsorted_key("baz")



import operator
from abc import abstractmethod
from collections.abc import Hashable, Mapping, MutableMapping, Sequence
from numbers import Number
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

import numpy as np
import pandas as pd
import pint

if TYPE_CHECKING:
    from genno.types import TQuantity, Unit

    from .quantity import AnyQuantity


class UnitsMixIn:
    """Object with :attr:`.units` and :meth:`._binary_op_units`."""

    attrs: dict[Hashable, Any]

    @property
    # def units(self) -> "Unit":
    def units(self):
        """Retrieve or set the units of the Quantity.

        Examples
        --------
        Create a quantity without units:

        >>> qty = Quantity(...)

        Set using a string; automatically converted to pint.Unit:

        >>> qty.units = "kg"
        >>> qty.units
        <Unit('kilogram')>

        """
        return self.attrs.setdefault(
            "_unit", pint.get_application_registry().dimensionless
        )

    @units.setter
    def units(self, value: "Unit | str") -> None:
        self.attrs["_unit"] = pint.get_application_registry().Unit(value)

    def _binary_op_units(
        self, other: "UnitsMixIn", op, swap: bool
    ) -> tuple["Unit", float]:
        """Determine result units for a binary operation between `self` and `other`.

        Returns:
        1. Result units.
        2. For rank-1 operations ('add', 'radd', 'rsub', 'sub') operations, a scaling
           factor to make magnitudes of `other` compatible with `self`.
        """
        # Retrieve units of `other`
        ou = other.units

        # Ensure there is not a mix of pint.Unit and pint.registry.Unit; this throws off
        # pint's internal logic
        if ou.__class__ is not self.units.__class__:
            ou = self.units.__class__(ou)

        if rank(op) == 1:
            # Determine multiplicative factor to align `other` to `self`
            return self.units, pint.Quantity(1.0, ou).to(self.units).magnitude
        elif rank(op) == 2:
            # Allow pint to determine the output units
            return op(*[ou, self.units] if swap else [self.units, ou]), np.nan
        else:
            # Exponent, its units, and base units
            exp, eu, bu = (self, self.units, ou) if swap else (other, ou, self.units)
            if not eu.dimensionless:
                raise ValueError(f"Cannot raise to a power with units {eu:~}")

            # Extract the (dense) data of the exponent
            data = cast("AnyQuantity", exp).to_series().values
            # Each exponent modulo 1. Set of {0} if exponents are all integers.
            check = set(np.mod(data, 1))
            # Unique values in data
            unique_values = np.unique(data)

            if check == {0.0} and len(unique_values) == 1:
                # The same, integer exponent for all values; raise the base units to
                # this value
                return op(bu, unique_values[0]), np.nan
            else:
                return pint.get_application_registry().dimensionless, np.nan


def make_binary_op(op, *, swap: bool):
    """Create a method for binary operator `name`."""

    def method(obj: "BaseQuantity", other: "BaseQuantity"):
        scalar_other = False
        if isinstance(other, Number):
            other = type(obj)(other)
            scalar_other = True
        elif not (
            isinstance(other, type(obj))
            or getattr(other, "__thisclass__", None) is type(obj)  # super()
        ):
            raise TypeError(type(other))

        left, right, result_units, factor = prepare_binary_op(obj, other, op, swap)

        # If `other` was scalar and the operation is rank-1 (add, sub, etc.), the units
        # of `obj` carry to the result. Otherwise, use `result_units`.
        return obj._keep(
            obj._perform_binary_op(op, left, right, factor),
            name=scalar_other,
            attrs=scalar_other,
            units=obj.units if (scalar_other and rank(op) == 1) else result_units,
        )

    return method


T = TypeVar("T")


class BinaryOpsMixIn(Generic[T]):
    """Binary operations for :class:`Quantity`.

    Subclasses **must** implement :meth:`_perform_binary_op`.

    Several binary operations are provided with methods that:

    - Convert scalar operands to :class:`.Quantity`.
    - Determine result units.
    - Preserve name and non-unit attrs.
    """

    __add__ = make_binary_op(operator.add, swap=False)
    __mul__ = make_binary_op(operator.mul, swap=False)
    __pow__ = make_binary_op(operator.pow, swap=False)
    __radd__ = make_binary_op(operator.add, swap=True)
    __rmul__ = make_binary_op(operator.mul, swap=True)
    __rpow__ = make_binary_op(operator.pow, swap=True)
    __rsub__ = make_binary_op(operator.sub, swap=True)
    __rtruediv__ = make_binary_op(operator.truediv, swap=True)
    __sub__ = make_binary_op(operator.sub, swap=False)
    __truediv__ = make_binary_op(operator.truediv, swap=False)

    @staticmethod
    @abstractmethod
    def _perform_binary_op(name: str, left: T, right: T, factor: float) -> T: ...


class BaseQuantity(
    BinaryOpsMixIn,
    UnitsMixIn,
):
    """Common base for a class that behaves like :class:`xarray.DataArray`.

    The class has units and unit-aware binary operations.
    """

    name: Hashable | None

    @abstractmethod
    def __init__(
        self,
        data: Any = None,
        coords: Sequence[tuple] | Mapping[Hashable, Any] | None = None,
        dims: str | Sequence[Hashable] | None = None,
        name: Hashable = None,
        attrs: Mapping | None = None,
        # internal parameters
        indexes: dict[Hashable, pd.Index] | None = None,
        fastpath: bool = False,
        **kwargs,
    ): ...

    def _keep(
        self,
        target: "TQuantity",
        attrs: Any | None = False,
        name: Any | None = False,
        units: Any | None = False,
    ) -> "TQuantity":
        """Preserve `name`, `units`, and/or other `attrs` from `self` to `target`.

        The action for each argument is:

        - :any:`False`: don't keep.
        - :any:`True`: keep the existing value.
        - Anything else: assign this value.
        """
        if name is not False:
            target.name = self.name if name is True else name
        if attrs is True:
            target.attrs.update(self.attrs)
        elif attrs is not False:
            assert isinstance(attrs, Mapping)
            target.attrs.update(attrs)
        if units is not False:
            # Only units; not other attrs
            target.units = self.units if units is True else units  # type: ignore [assignment]
        return target


def prepare_binary_op(
    obj: BaseQuantity, other, op, swap: bool
) -> tuple[BaseQuantity, BaseQuantity, "Unit", float]:
    """Prepare inputs for a binary operation.

    Returns:

    1. The left operand (`obj` if `swap` is False else `other`).
    2. The right operand. If units of `other` are different than `obj`, `other` is
       scaled.
    3. Units for the result. In additive operations, the units of `obj` take precedence.
    4. Any scaling factor needed to make units of `other` compatible with `obj`.
    """
    # Determine resulting units
    result_units, factor = obj._binary_op_units(other, op, swap)

    # Apply a multiplicative factor to align units
    if rank(op) == 1 and factor != 1.0:
        other = super(type(obj), other).__mul__(factor)

    # For __r*__ methods
    left, right = (other, obj) if swap else (obj, other)

    return left, right, result_units, factor


def collect_attrs(
    data, attrs_arg: Mapping | None, kwargs: MutableMapping
) -> MutableMapping:
    """Handle `attrs` and 'units' `kwargs` to Quantity constructors."""
    # Use attrs, if any, from an existing object, if any
    new_attrs = getattr(data, "attrs", dict()).copy()

    # Overwrite with values from an explicit attrs argument
    new_attrs.update(attrs_arg or dict())

    # Store the "units" keyword argument as an attr
    units = kwargs.pop("units", None)
    if units is not None:
        new_attrs["_unit"] = pint.Unit(units)

    return new_attrs


def rank(op) -> int:
    """Rank of the binary operation `op`.

    See `‘Hyperoperation’ on Wikipedia <https://en.wikipedia.org/wiki/Hyperoperation>`_
    for the sense of this meaning of ‘rank’.
    """
    return {
        operator.add: 1,
        operator.sub: 1,
        operator.mul: 2,
        operator.truediv: 2,
        operator.pow: 3,
    }[op]


def single_column_df(data, name: Hashable) -> tuple[Any, Hashable]:
    """Handle `data` and `name` arguments to Quantity constructors."""
    if isinstance(data, pd.DataFrame):
        if len(data.columns) != 1:
            raise TypeError(
                f"Cannot instantiate Quantity from {len(data.columns)}-D data frame"
            )

        # Unpack a single column; use its name if not overridden by `name`
        return data.iloc[:, 0], (name or data.columns[0])
    else:
        return data, name



from importlib.util import find_spec

#: :class:`bool` indicating whether :mod:`plotnine` is available.
HAS_PLOTNINE = find_spec("plotnine") is not None

if HAS_PLOTNINE:
    from .plot import Plot

    __all__ = ["Plot"]



import logging
import re
from collections.abc import Callable, Generator, Hashable, Iterable, Iterator, Sequence
from functools import partial, singledispatch
from itertools import chain, compress
from types import MappingProxyType
from typing import TYPE_CHECKING, SupportsInt
from warnings import warn

from .attrseries import AttrSeries
from .sparsedataarray import SparseDataArray

if TYPE_CHECKING:
    from .quantity import AnyQuantity


log = logging.getLogger(__name__)

#: Regular expression for valid key strings.
EXPR = re.compile(r"^(?P<name>[^:]+)(:(?P<dims>([^:-]*-)*[^:-]+)?(:(?P<tag>[^:]*))?)?$")

#: Regular expression for non-keylike strings.
BARE_STR = re.compile(r"^\s*(?P<name>[^:]+)\s*$")


@singledispatch
def _name_dims_tag(value) -> tuple[str, tuple[str, ...], str | None]:
    """Convert various `value`s into (name, dims, tag) tuples.

    Helper for :meth:`.Key.__init__`.
    """
    raise TypeError(type(value))


@_name_dims_tag.register
def _(value: str):
    """Parse a string that may contain a Key expression."""
    match = EXPR.match(value)
    if match is None:
        raise ValueError(f"Invalid key expression: {repr(value)}")
    groups = match.groupdict()
    return (
        groups["name"],
        tuple() if not groups["dims"] else tuple(groups["dims"].split("-")),
        groups["tag"],
    )


@_name_dims_tag.register(AttrSeries)
@_name_dims_tag.register(SparseDataArray)
def _(value: "AnyQuantity"):  # register() only handles bare AnyQuantity in Python ≥3.11
    """Return (name, dims, tag) that describe an existing Quantity."""
    return str(value.name), tuple(map(str, value.dims)), None


class KeyGeneratorMixIn:
    """Mix-in class for classes that can derive :class:`.Key` from a base."""

    __slots__ = ("_base", "_generated")

    _base: "Key"
    _generated: list[Hashable]

    def __init__(self) -> None:
        self._generated = []

    def __call__(self, value: Hashable | None = None) -> "Key":
        return next(self) if value is None else self[value]

    def __getitem__(self, value: Hashable) -> "Key":
        value = int(value) if isinstance(value, SupportsInt) else str(value)
        if value not in self._generated:
            self._generated.append(value)
        return self._base.add_tag(str(value))

    def __next__(self) -> "Key":
        return self[self._next_int_tag()]

    def _next_int_tag(self) -> int:
        return max([-1] + [t for t in self._generated if isinstance(t, int)]) + 1

    @property
    def generated(self) -> tuple["Key", ...]:
        """Sequence of previously-created :class:`Keys <.Key>`."""
        return tuple(self._base.add_tag(str(k)) for k in self._generated)

    @property
    def last(self) -> "Key":
        """The most recently created :class:`.Key`."""
        return self._base.add_tag(str(self._generated[-1]))


class Key(KeyGeneratorMixIn):
    """A hashable key for a quantity that includes its dimensionality."""

    __slots__ = ("_dims", "_hash", "_name", "_str", "_tag")

    _dims: tuple[str, ...]
    _hash: int
    _name: str
    _str: str
    _tag: str | None

    def __init__(
        self,
        name_or_value: "str | Key | AnyQuantity",
        dims: Iterable[str] = [],
        tag: str | None = None,
        _fast: bool = False,
    ):
        if _fast:
            # Fast path: don't handle arguments
            assert isinstance(name_or_value, str)
            self._name = name_or_value
            self._dims = tuple(dims)
            self._tag = tag or None
        else:
            # Convert various values into a (name, dims, tags)
            self._name, _dims, _tag = _name_dims_tag(name_or_value)

            # Check for conflicts between dims inferred from name_or_value and any
            # direct argument
            # TODO handle resolveable combinations without raising exceptions
            if bool(_dims) and bool(dims):
                raise ValueError(
                    f"Conflict: {dims = } argument vs. {_dims!r} from {name_or_value!r}"
                )
            elif bool(_tag) and bool(tag):
                raise ValueError(
                    f"Conflict: {tag = } argument vs. {_tag!r} from {name_or_value!r}"
                )

            self._dims = _dims or tuple(dims)
            self._tag = _tag or tag

        super().__init__()
        self._base = self

        # Pre-compute string representation and hash
        self._str = (
            self._name
            + ":"
            + "-".join(self._dims)
            + (f":{self._tag}" if self._tag else "")
        )
        # Hash is independent of dim order
        self._hash = hash(
            self._name
            + ":"
            + "-".join(sorted(self._dims))
            + (f":{self._tag}" if self._tag else "")
        )

    # Class methods

    @classmethod
    def bare_name(cls, value) -> str | None:
        """If `value` is a bare name (no dims or tags), return it; else :obj:`None`."""
        if not isinstance(value, str):
            return None
        match = BARE_STR.match(value)
        return match.group("name") if match else None

    @classmethod
    def from_str_or_key(
        cls,
        value: "str | Key | AnyQuantity",
        drop: Iterable[str] | bool = [],
        append: Iterable[str] = [],
        tag: str | None = None,
    ) -> "Key":
        """Return a new Key from *value*.

        .. versionchanged:: 1.18.0

           Calling :meth:`from_str_or_key` with a single argument is no longer
           necessary; simply give the same `value` as an argument to :class:`Key`.

           The class method is retained for convenience when calling with multiple
           arguments. However, the following are equivalent and may be more readable:

           .. code-block:: python

              k1 = Key("foo:a-b-c:t1", drop="b", append="d", tag="t2")
              k2 = Key("foo:a-b-c:t1").drop("b").append("d)"

        Parameters
        ----------
        value : str or .Key
            Value to use to generate a new Key.
        drop : list of str or :obj:`True`, optional
            Existing dimensions of *value* to drop. See :meth:`drop`.
        append : list of str, optional
            New dimensions to append to the returned Key. See :meth:`append`.
        tag : str, optional
            Tag for returned Key. If *value* has a tag, the two are joined
            using a '+' character. See :meth:`add_tag`.

        Returns
        -------
        :class:`Key`
        """
        base = cls(value)

        # Return quickly if no further manipulations are required
        if not any([drop, append, tag]):
            warn(
                "Calling Key.from_str_or_key(value) with no other arguments is no "
                "longer necessary; simply use Key(value)",
                FutureWarning,
                stacklevel=2,
            )
            return base

        # mypy is fussy here
        drop_args: tuple[str | bool, ...] = tuple(
            [drop] if isinstance(drop, bool) else drop
        )

        # Drop and append dimensions; add tag
        return base.drop(*drop_args).append(*tuple(append)).add_tag(tag)

    @classmethod
    def product(cls, new_name: str, *keys, tag: str | None = None) -> "Key":
        """Return a new Key that has the union of dimensions on *keys*.

        Dimensions are ordered by their first appearance:

        1. First, the dimensions of the first of the *keys*.
        2. Next, any additional dimensions in the second of the *keys* that
           were not already added in step 1.
        3. etc.

        Parameters
        ----------
        new_name : str
            Name for the new Key. The names of *keys* are discarded.
        keys
            May include instances of :class:`.Key`, :class:`str` (converted to Key), or
            :class:`Quantity` (the dimensions of the quantity are used directly).
        """
        # Iterable of dimension names from all keys, in order, with repetitions
        dims = chain(
            *map(
                lambda k: cls(k).dims
                if isinstance(k, (AttrSeries, SparseDataArray, Key, str))
                else (),
                keys,
            )
        )

        # Return new key. Use dict to keep only unique *dims*, in same order
        return cls(new_name, dict.fromkeys(dims).keys()).add_tag(tag)

    def __add__(self, other: str) -> "Key":
        if not isinstance(other, str):
            raise TypeError(type(other))
        return self.add_tag(other)

    def __sub__(self, other: str | Iterable[str]) -> "Key":
        return self.remove_tag(*((other,) if isinstance(other, str) else other))

    def __mul__(self, other: "str | Key | Sequence[str]") -> "Key":
        if isinstance(other, str):
            other_dims: Sequence[str] = (other,)
        elif isinstance(other, Key):
            other_dims = other.dims
        elif isinstance(other, Sequence):
            other_dims = other
        else:
            raise TypeError(type(other))

        return self.append(*other_dims)

    def __truediv__(self, other: "str | Key | Sequence[str]") -> "Key":
        if isinstance(other, str):
            other_dims: Sequence[str] = (other,)
        elif isinstance(other, Key):
            other_dims = other.dims
        elif isinstance(other, Sequence):
            other_dims = other
        else:
            raise TypeError(type(other))

        return self.drop(*other_dims)

    def __repr__(self) -> str:
        """Representation of the Key, e.g. '<name:dim1-dim2-dim3:tag>."""
        return f"<{self._str}>"

    def __str__(self) -> str:
        """String equivalent of the Key, e.g. 'name:dim1-dim2-dim3:tag'."""
        return self._str  # Return the pre-computed value

    def __hash__(self):
        """Key hashes the same as :py:`str(Key)`."""
        return self._hash

    def __eq__(self, other) -> bool:
        """Key is equal to :py:`str(Key)`."""
        try:
            other = Key(other)
        except TypeError:
            return NotImplemented

        return (
            (self.name == other.name)
            and (set(self.dims) == set(other.dims))
            and (self.tag == other.tag)
        )

    # Less-than and greater-than operations, for sorting
    def __lt__(self, other) -> bool:
        if isinstance(other, Key):
            return str(self.sorted) < str(other.sorted)
        elif isinstance(other, str):
            return str(self.sorted) < other
        else:
            return NotImplemented

    def __gt__(self, other) -> bool:
        if isinstance(other, Key):
            return str(self.sorted) > str(other.sorted)
        elif isinstance(other, str):
            return str(self.sorted) > other
        else:
            return NotImplemented

    @property
    def name(self) -> str:
        """Name of the quantity, :class:`str`."""
        return self._name

    @property
    def dims(self) -> tuple[str, ...]:
        """Dimensions of the quantity, :class:`tuple` of :class:`str`."""
        return self._dims

    @property
    def tag(self) -> str | None:
        """Quantity tag, :class:`str` or :obj:`None`."""
        return self._tag

    @property
    def sorted(self) -> "Key":
        """A version of the Key with its :attr:`.dims` :func:`sorted`."""
        return Key(self._name, sorted(self._dims), self._tag, _fast=True)

    def rename(self, name: str) -> "Key":
        """Return a Key with a replaced `name`."""
        return Key(name, self._dims, self._tag, _fast=True)

    def drop(self, *dims: str | bool) -> "Key":
        """Return a new Key with `dims` dropped."""
        return Key(
            self._name,
            tuple() if dims == (True,) else filter(lambda d: d not in dims, self._dims),
            self._tag,
            _fast=True,
        )

    def drop_all(self) -> "Key":
        """Return a new Key with all dimensions dropped / zero dimensions."""
        return Key(self._name, tuple(), self._tag, _fast=True)

    def append(self, *dims: str) -> "Key":
        """Return a new Key with additional dimensions `dims`."""
        return Key(self._name, list(self._dims) + list(dims), self._tag, _fast=True)

    def add_tag(self, tag: str | None) -> "Key":
        """Return a new Key with `tag` appended."""
        return Key(
            self._name, self._dims, "+".join(filter(None, [self._tag, tag])), _fast=True
        )

    def iter_sums(self) -> Generator[tuple["Key", Callable, "Key"], None, None]:
        """Generate (key, task) for all possible partial sums of the Key."""
        from genno.operator import sum

        for agg_dims, others in combo_partition(self.dims):
            yield (
                Key(self._name, agg_dims, self.tag, _fast=True),
                partial(sum, dimensions=others, weights=None),
                self,
            )

    def remove_tag(self, *tags: str) -> "Key":
        """Return a key with any of `tags` dropped.

        Raises
        ------
        ValueError
            If none of `tags` are in :attr:`.tags`.
        """
        new_tags = tuple(filter(lambda t: t not in tags, (self.tag or "").split("+")))
        new_tag = "+".join(new_tags) if new_tags else None
        if new_tag == self.tag:
            raise ValueError(f"No existing tags {tags!r} to remove")
        return Key(self._name, self._dims, new_tag, _fast=True)


@_name_dims_tag.register
def _(value: Key):
    """Return the (name, dims, tag) of an existing Key."""
    return value._name, value._dims, value._tag


class Keys:
    """A collection of :class:`.Key`.

    This is essentially the same as :class:`.types.SimpleNamespace`, except every
    attribute is a :class:`.Key`.
    """

    __slots__ = ("_keys",)

    _keys: dict[str, Key]

    def __init__(self, **kwargs: "KeyLike") -> None:
        object.__setattr__(self, "_keys", {})
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __delattr__(self, name: str) -> None:
        self._keys.pop(name)

    def __getattr__(self, name: str) -> "Key":
        try:
            return self._keys[name]
        except KeyError:
            raise AttributeError(name) from None

    def __repr__(self) -> str:
        return f"<{len(self._keys)} keys: {' '.join(sorted(self._keys))}>"

    def __setattr__(self, name: str, value: "Key") -> None:
        self._keys[name] = value if isinstance(value, Key) else Key(value)


class KeySeq(KeyGeneratorMixIn):
    """Utility class for generating similar :class:`Keys <.Key>`."""

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._base = Key(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<KeySeq from '{self._base!s}'>"

    # Particular to KeySeq

    @property
    def keys(self) -> MappingProxyType:
        """Read-only view of previously-created :class:`Keys <.Key>`.

        In the form of a :class:`dict` mapping tags (:class:`int` or :class:`str`) to
        :class:`.Key` values.
        """
        return MappingProxyType(
            {k: self._base.add_tag(str(k)) for k in self._generated}
        )

    @property
    def prev(self) -> Key:
        """Alias of :attr:`.KeyGeneratorMixin.last`."""
        return self.last

    # Access to Key properties
    @property
    def base(self) -> Key:
        """The base Key."""
        return self._base

    @property
    def name(self) -> str:
        """Name of the :attr:`.base` Key."""
        return self._base.name

    @property
    def dims(self) -> tuple[str, ...]:
        """Dimensions of the :attr:`.base` Key."""
        return self._base.dims

    @property
    def tag(self) -> str | None:
        """Tag of the :attr:`.base` Key."""
        return self._base.tag

    def __add__(self, other: str) -> "KeySeq":
        return KeySeq(self._base.__add__(other))

    def __mul__(self, other) -> "KeySeq":
        return KeySeq(self._base.__mul__(other))

    def __sub__(self, other: str | Iterable[str]) -> "KeySeq":
        return KeySeq(self._base.__sub__(other))

    def __truediv__(self, other) -> "KeySeq":
        return KeySeq(self._base.__truediv__(other))


#: Type shorthand for :class:`Key` or any other value that can be used as a key.
KeyLike = Key | str


def combo_partition(iterable):
    """Yield pairs of lists with all possible subsets of *iterable*."""
    # Format string for binary conversion, e.g. '04b'
    fmt = "0{}b".format(len(iterable))
    for n in range(2 ** len(iterable) - 1):
        # Two binary lists
        a, b = zip(*[(v, not v) for v in map(int, format(n, fmt))])
        yield list(compress(iterable, a)), list(compress(iterable, b))


def iter_keys(value: KeyLike | tuple[KeyLike, ...]) -> Iterator[Key]:
    """Yield :class:`Keys <Key>` from `value`.

    Raises
    ------
    TypeError
        `value` is not an iterable of :class:`Key`.

    See also
    --------
    .Computer.add
    """
    if isinstance(value, (Key, str)):
        yield Key(value)
        tmp: Iterator[KeyLike] = iter(())
    else:
        tmp = iter(value)
    for element in tmp:
        if not isinstance(element, Key):
            raise TypeError(type(element))
        yield element


def single_key(value: KeyLike | tuple[KeyLike, ...] | Iterator) -> Key:
    """Ensure `value` is a single :class:`Key`.

    Raises
    ------
    TypeError
        `value` is not a :class:`Key` or 1-tuple of :class:`Key`.

    See also
    --------
    .Computer.add
    """
    if isinstance(value, (Key, str)):
        return Key(value)

    tmp = iter(value)
    try:
        result = next(tmp)
    except StopIteration:
        raise TypeError("Empty iterable")
    else:
        try:
            next(tmp)
        except StopIteration:
            pass
        else:
            raise TypeError("Iterable of length >1")

    if isinstance(result, Key):
        return result
    else:
        raise TypeError(type(result))
